#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Tests for dynamically defined operations.

Test the command-line wrapping functionality in gmxapi.commandline. These
operation factories are written with user-facing tools and exercise a lot of
the high-level machinery of the package, effectively serving as integration
tests of the operation-building utilities in the modules depended on by
commandline.py.
"""
import logging
import os
import shutil
import stat
import tempfile
from pathlib import Path

import pytest

import gmxapi as gmx
import gmxapi.operation
from gmxapi import commandline
from gmxapi.testsupport import scoped_chdir

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    rank_number = comm.Get_rank()
    comm_size = comm.Get_size()
except ImportError:
    comm = None
    rank_number = 0
    comm_size = 1
    rank_tag = ""
    MPI = None
else:
    rank_tag = "rank{}:".format(rank_number)


# if rank_number == 1:
#     import pydevd_pycharm
#     pydevd_pycharm.settrace('localhost', port=12345, stdoutToServer=True, stderrToServer=True)


def test_true_base(cleandir):
    """Test a command known to produce a return code of 0."""
    work_dir = cleandir
    if comm_size > 1:
        work_dir = comm.bcast(work_dir, root=0)
    with scoped_chdir(work_dir):
        command = shutil.which("true")
        kwargs = dict(command=[command], shell=False)
        if gmx.version.api_is_at_least(0, 4):
            kwargs["env"] = dict(os.environ)
        operation = commandline.cli(**kwargs)

        # Note: getitem not implemented.
        # assert 'stdout' in operation.output
        # assert 'stderr' in operation.output
        assert hasattr(operation.output, "stdout")
        assert hasattr(operation.output, "stderr")
        assert not hasattr(operation.output, "erroroutput")
        assert hasattr(operation.output, "returncode")

        operation.run()
        # assert operation.output.returncode.result() == 0
        assert operation.output.returncode.result() == 0


def test_false_explicit(cleandir):
    """Test a command known to produce a return code of 1."""
    command = shutil.which("false")
    kwargs = dict(command=[command], shell=False)
    if gmx.version.api_is_at_least(0, 4):
        kwargs["env"] = dict(os.environ)
    operation = commandline.cli(**kwargs)  # Explicitly run the operation.
    operation.run()
    assert operation.output.returncode.result() == 1


def test_false_implicit(cleandir):
    command = shutil.which("false")
    kwargs = dict(command=[command], shell=False)
    if gmx.version.api_is_at_least(0, 4):
        kwargs["env"] = dict(os.environ)
    operation = commandline.cli(**kwargs)
    # Allow the operation to be executed implicitly to satisfy data constraint.
    assert operation.output.returncode.result() == 1


def test_command_with_arguments(cleandir):
    """Test that cli() can wrap a command with arguments."""
    kwargs = dict(command=[shutil.which("echo"), "hi", "there"], shell=False)
    if gmx.version.api_is_at_least(0, 4):
        kwargs["env"] = dict(os.environ)
    operation = commandline.cli(**kwargs)
    assert operation.output.returncode.result() == 0


def test_command_with_input_files(cleandir):
    """Test handling of *input_file* arguments.

    Generate a file and provide it as a relative path to a simple command line tool.
    Confirm that the relative path (valid for the script working directory but not
    for the Task working directory) is correctly normalized to an absolute path.

    Check for #4827 regression.
    """
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as fh:
        fh.write("hi\nthere\n")
        fh.flush()
    assert fh.closed
    actual_path = Path(fh.name).resolve()
    try:
        with open(actual_path, "r") as fh:
            num_lines = len(fh.readlines())
        relative_path = os.path.relpath(actual_path, start=cleandir)
        # Overload the input files facility to trigger path normalization.
        # `wc` implementations are fairly consistent as far as providing a
        # `-l` option and accepting a positional file argument.
        cmd = commandline.commandline_operation(
            executable="wc", input_files={"-l": relative_path}
        )
        # We want to find a relative path from the current directory that would
        # not resolve correctly from the generated Task directory. We will now
        # double-check that the relative path _would have_ caused the test to fail
        # if the absolute path conversion (#4827) were not working correctly.
        workdir = cmd.output.directory.result()
        assert not Path(workdir).joinpath(relative_path).exists()
        # Uncomment the following for additional debugging.
        # actual_relative_path = os.path.relpath(actual_path, start=workdir)
        # returncode = cmd.output.returncode.result()
        # error = cmd.output.stderr.result()

        # Check that the file whose lines we counted has the expected number of lines.
        output = cmd.output.stdout.result()
        # For `wc -l filename` on a 2-line file, we expect output
        # similar to "      2 filename\n".
        # Check the first non-whitespace field.
        assert output.split()[0] == str(num_lines)
    finally:
        os.unlink(actual_path)


def test_command_with_stdin(cleandir):
    """Test that cli() can handle string input."""
    stdin = "hi\nthere\n"
    subcommand = "{wc} -l | {grep} -q 2".format(
        wc=shutil.which("wc"), grep=shutil.which("grep")
    )

    kwargs = dict(command=["/bin/sh", "-c", subcommand], shell=False, stdin=stdin)
    if gmx.version.api_is_at_least(0, 4):
        kwargs["env"] = dict(os.environ)
    operation = commandline.cli(**kwargs)
    assert operation.output.returncode.result() == 0
    operation = commandline.commandline_operation(
        "/bin/sh", ["-c", subcommand], stdin=stdin
    )
    assert operation.output.returncode.result() == 0

    subcommand = "{wc} -l | {grep} -q 1".format(
        wc=shutil.which("wc"), grep=shutil.which("grep")
    )

    kwargs = dict(command=["/bin/sh", "-c", subcommand], shell=False, stdin=stdin)
    if gmx.version.api_is_at_least(0, 4):
        kwargs["env"] = dict(os.environ)
    operation = commandline.cli(**kwargs)
    assert operation.output.returncode.result() != 0
    operation = commandline.commandline_operation(
        "/bin/sh", ["-c", subcommand], stdin=stdin
    )
    assert operation.output.returncode.result() != 0


def test_true(cleandir):
    operation = commandline.commandline_operation(executable="true")
    # Note: getitem not implemented.
    # assert 'stdout' in operation.output
    # assert 'stderr' in operation.output
    assert not hasattr(operation.output, "erroroutput")
    assert hasattr(operation.output, "file")
    assert hasattr(operation.output, "stdout")
    assert hasattr(operation.output, "stderr")
    assert hasattr(operation.output, "returncode")
    assert operation.output.returncode.result() == 0


def test_false(cleandir):
    operation = commandline.commandline_operation(executable="false")
    assert operation.output.returncode.result() == 1


def test_echo(cleandir):
    operation = commandline.commandline_operation(
        executable="echo", arguments=["hi there"]
    )
    assert operation.output.stdout.result() == "hi there\n"
    assert operation.output.returncode.result() == 0


def test_file_dependency_chain(cleandir):
    """Test the command line wrapper input/output file handling.

    Operation output can be used as operation input.
    """
    file1 = os.path.join(cleandir, "input")
    file2 = os.path.join(cleandir, "output")

    # Make a shell script that acts like the type of tool we are wrapping.
    scriptname = os.path.join(cleandir, "clicommand.sh")
    with open(scriptname, "w") as fh:
        fh.write(
            "\n".join(
                [
                    "#!" + shutil.which("bash"),
                    "# Concatenate an input file and a string argument to an output file.",
                    "# Mock a utility with the tested syntax.",
                    '#     clicommand.sh "some words" -i inputfile -o outputfile',
                    "echo $1 | cat $3 - > $5\n",
                ]
            )
        )
    os.chmod(scriptname, stat.S_IRWXU)

    line1 = "first line"
    filewriter1 = gmx.commandline_operation(
        scriptname,
        arguments=[line1],
        input_files={"-i": os.devnull},
        output_files={"-o": file1},
    )
    assert isinstance(filewriter1.output, gmxapi.operation.DataProxyBase)
    assert filewriter1.output.ensemble_width == 1

    line2 = "second line"
    filewriter2 = gmx.commandline_operation(
        scriptname,
        arguments=[line2],
        input_files={"-i": filewriter1.output.file["-o"]},
        output_files={"-o": file2},
    )

    filewriter2.run()
    # Check that the files have the expected lines
    with open(filewriter1.output.file["-o"].result(), "r") as fh:
        lines = [text.rstrip() for text in fh]
    assert len(lines) == 1
    assert lines[0] == line1
    with open(filewriter2.output.file["-o"].result(), "r") as fh:
        lines = [text.rstrip() for text in fh]
    assert len(lines) == 2
    assert lines[0] == line1
    assert lines[1] == line2

    # Make sure that the temporary directory is not removed before all ranks have done
    # the file checks.
    if comm_size > 1:
        comm.barrier()


def test_file_dependency_ensemble_chain(cleandir):
    """Test the command line wrapper input/output file handling.

    Operation output can be used as operation input.
    """
    file1 = (os.path.join(cleandir, "input1"), os.path.join(cleandir, "input2"))
    file2 = "output"

    # Make a shell script that acts like the type of tool we are wrapping.
    scriptname = os.path.join(cleandir, "clicommand.sh")
    with open(scriptname, "w") as fh:
        fh.write(
            "\n".join(
                [
                    "#!" + shutil.which("bash"),
                    "# Concatenate an input file and a string argument to an output file.",
                    "# Mock a utility with the tested syntax.",
                    '#     clicommand.sh "some words" -i inputfile -o outputfile',
                    "echo $1 | cat $3 - > $5\n",
                ]
            )
        )
    os.chmod(scriptname, stat.S_IRWXU)

    line1 = ["first line A", "first line B"]
    filewriter1 = gmx.commandline_operation(
        scriptname,
        arguments=[[line] for line in line1],
        input_files={"-i": os.devnull},
        output_files=[{"-o": name} for name in file1],
    )
    assert isinstance(filewriter1.output, gmxapi.operation.DataProxyBase)
    assert filewriter1.output.ensemble_width == 2

    line2 = "second line"
    filewriter2 = gmx.commandline_operation(
        scriptname,
        arguments=[line2],
        input_files={"-i": filewriter1.output.file["-o"]},
        output_files={"-o": file2},
    )
    assert filewriter2.output.ensemble_width == 2
    filewriter2.run()

    # Check that the files have the expected lines
    for member in range(2):
        with open(filewriter1.output.file["-o"].result()[member], "r") as fh:
            lines = [text.rstrip() for text in fh]
        assert len(lines) == 1
        assert lines[0] == line1[member]

    outputs = filewriter2.output.file["-o"].result()
    assert len(outputs) == 2
    assert outputs[0] != outputs[1]
    for member in range(2):
        path = outputs[member]
        assert os.path.exists(path)
        with open(path, "r") as fh:
            lines = [text.rstrip() for text in fh]
            assert len(lines) == 2
            assert lines[0] == line1[member]
            assert lines[1] == line2
    # Make sure that the temporary directory is not removed before all ranks have done
    # the file checks.
    if comm_size > 1:
        comm.barrier()


def test_failure(cleandir):
    """The operation should not deliver file output if the subprocess fails."""
    file1 = os.path.join(cleandir, "input")
    file2 = os.path.join(cleandir, "output")

    # Make a shell script that acts like the type of tool we are wrapping.
    scriptname = os.path.join(cleandir, "clicommand.sh")
    with open(scriptname, "w") as fh:
        fh.write(
            "\n".join(
                [
                    "#!" + shutil.which("bash"),
                    "# Concatenate an input file and a string argument to an output file.",
                    "# Mock a utility with the tested syntax.",
                    '#     clicommand.sh "some words" -i inputfile -o outputfile',
                    "exit 1\n",
                ]
            )
        )
    os.chmod(scriptname, stat.S_IRWXU)

    filewriter1 = gmx.commandline_operation(
        scriptname, input_files={"-i": os.devnull}, output_files={"-o": file1}
    )

    filewriter2 = gmx.commandline_operation(
        scriptname,
        input_files={"-i": filewriter1.output.file["-o"]},
        output_files={"-o": file2},
    )

    with pytest.warns(UserWarning, match=r"Trouble getting .*"):
        # We expect this to generate errors, warnings, and log messages.
        # We assert the exception and warning, and suppress the logged error
        # to avoid misleading output from the test suite.

        class _Filter(logging.Filter):
            def filter(self, record: logging.LogRecord):
                if (
                    record.name == "gmxapi.operation"
                    and "Trouble getting" in record.msg
                ):
                    return 0
                else:
                    return 1

        _filter = _Filter()
        logging.getLogger("gmxapi.operation").addFilter(_filter)

        with pytest.raises(KeyError):
            filewriter2.run()

        logging.getLogger("gmxapi.operation").removeFilter(_filter)



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2021- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Test gmxapi._gmxapi file I/O routines."""
import os
import tempfile

import pytest
import gmxapi
import gmxapi.simulation.fileio
from gmxapi import _gmxapi


@pytest.mark.usefixtures("cleandir")
def test_core_read_tpr(spc_water_box):
    tpr_filehandle: _gmxapi.TprFile = _gmxapi.read_tprfile(os.fsencode(spc_water_box))
    parameters: _gmxapi.SimulationParameters = tpr_filehandle.params()
    assert "nsteps" in parameters.extract()
    assert "foo" not in parameters.extract()
    assert parameters.extract()["nsteps"] == 2


@pytest.mark.usefixtures("cleandir")
def test_core_rewrite_tprfile(spc_water_box):
    """Test _gmxapi.rewrite_tprfile for update of end_time.

    Set a new end time that is 5000 steps later than the original. Read dt
    from file to avoid floating point round-off errors.

    Transitively test gmx.fileio.read_tpr()
    """
    tpr_filename = spc_water_box
    additional_steps = 5000
    tpr_filehandle: _gmxapi.TprFile = _gmxapi.read_tprfile(os.fsencode(spc_water_box))
    params = tpr_filehandle.params().extract()
    dt = params["dt"]
    nsteps = params["nsteps"]
    init_step = params["init-step"]
    initial_endtime = (init_step + nsteps) * dt
    new_endtime = initial_endtime + additional_steps * dt
    _, temp_filename = tempfile.mkstemp(suffix=".tpr")
    _gmxapi.rewrite_tprfile(
        source=tpr_filename, destination=temp_filename, end_time=new_endtime
    )
    tprfile = gmxapi.simulation.fileio.TprFile(temp_filename, "r")
    with tprfile as fh:
        params = gmxapi.simulation.fileio.read_tpr(fh).parameters.extract()
        dt = params["dt"]
        nsteps = params["nsteps"]
        init_step = params["init-step"]
        assert (init_step + nsteps) * dt == new_endtime

    os.unlink(temp_filename)


@pytest.mark.usefixtures("cleandir")
def test_core_read_and_write_tpr_file(spc_water_box):
    """Test gmx.fileio.write_tpr_file() using gmx.core API."""
    additional_steps = 5000

    tpr_filehandle: _gmxapi.TprFile = _gmxapi.read_tprfile(os.fsencode(spc_water_box))
    sim_input: _gmxapi.SimulationParameters = tpr_filehandle.params()
    params: dict = sim_input.extract()
    nsteps = params["nsteps"]
    init_step = params["init-step"]

    # Choose a new nsteps to check integer parameter setting.
    new_nsteps = init_step + additional_steps
    # Choose a new dt to check floating point parameter setting
    new_dt = params["dt"] * 2.0

    sim_input.set("nsteps", new_nsteps)
    sim_input.set("dt", new_dt)

    _, temp_filename = tempfile.mkstemp(suffix=".tpr")
    _gmxapi.write_tprfile(temp_filename, sim_input)

    tprfile = gmxapi.simulation.fileio.TprFile(temp_filename, "r")
    with tprfile as fh:
        params = gmxapi.simulation.fileio.read_tpr(fh).parameters.extract()
        # Note that we have chosen an exactly representable dt for spc_water_box.
        # Otherwise, we would have to use pytest.approx with a suitable tolerance.
        assert params["dt"] == new_dt
        assert params["nsteps"] != nsteps
        assert params["nsteps"] == new_nsteps

    os.unlink(temp_filename)



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""modify_input operation module

Provides implementation classes and user interface for gmxapi.modify_input.

The *modify_input* operation accepts simulation input, such as from *read_tpr*,
and allows aspects of the simulation input to be overridden.
"""

__all__ = ["modify_input"]

import collections.abc
import inspect
import typing

import gmxapi
import gmxapi.abc
import gmxapi.exceptions
import gmxapi.operation as _op
import gmxapi.typing
from gmxapi.operation import InputCollectionDescription, ResourceManager
from .abc import ModuleObject

# Initialize module-level logger
from gmxapi import logger as root_logger

logger = root_logger.getChild("modify_input")
logger.info("Importing {}".format(__name__))

_output_descriptors = (
    _op.OutputDataDescriptor("_simulation_input", str),
    _op.OutputDataDescriptor("parameters", dict),
)
_publishing_descriptors = {
    desc._name: gmxapi.operation.Publisher(desc._name, desc._dtype)
    for desc in _output_descriptors
}
_output = _op.OutputCollectionDescription(
    **{descriptor._name: descriptor._dtype for descriptor in _output_descriptors}
)


class OutputDataProxy(ModuleObject, _op.DataProxyBase, descriptors=_output_descriptors):
    """Implement the 'output' attribute of `modify_input` operations.

    Attributes:
        parameters: Aggregated dictionary simulation parameters for the resulting simulation input.

    Additionally (through an unspecified interface), the object serves as a
    complete simulation input to other gmxapi operations.
    """

    def __init__(self, *args, **kwargs):
        _op.DataProxyBase.__init__(self, *args, **kwargs)


class PublishingDataProxy(_op.DataProxyBase, descriptors=_publishing_descriptors):
    """Manage output resource updates for ReadTpr operation."""


_output_factory = _op.OutputFactory(
    output_proxy=OutputDataProxy,
    output_description=_output,
    publishing_data_proxy=PublishingDataProxy,
)


class SessionResources(object):
    """Input and output run-time resources for a ModifyTPR operation."""

    def __init__(
        self, _simulation_input: str, parameters: dict, publisher: PublishingDataProxy
    ):
        """Initialize resources for a gmxapi.operation context.

        Arguments:
            parameters: Source of new simulation parameters data.
            publisher: Output publishing resource provided by the Context.
            _simulation_input: Unspecified. Reserved for implementation details.

        """
        # For the initial implementation, _simulation_input is just a string that
        # will be interpreted as a TPR file path. The typing will change in the
        # near term as we build the data flow model. Before 1.0, unusual members
        # like this will be removed or hidden in the interface of the other
        # input objects.
        assert isinstance(_simulation_input, str)
        self._simulation_input = _simulation_input
        assert isinstance(parameters, collections.abc.Mapping)
        self.parameters = parameters
        self.output = publisher


#
# Helpers
#

# The '_simulation_input' input is a workaround until the data model is improved and the
# simulation module is more integrated.
# Note that when a read_tpr handle is passed with the "input" key word argument,
# its outputs will be mapped directly to modify_input inputs. The presence of
# a user-supplied 'parameters' argument overrides the 'parameters' value from 'input'.
# The lack of a default on 'parameters' makes it required.
_input = _op.InputCollectionDescription(
    [
        (
            "_simulation_input",
            inspect.Parameter(
                "_simulation_input",
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=str,
            ),
        ),
        (
            "parameters",
            inspect.Parameter(
                "parameters", inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=dict
            ),
        ),
    ]
)


def _session_resource_factory(
    input: _op.InputPack, output: PublishingDataProxy, **kwargs
) -> SessionResources:
    """Translate resources from the gmxapi.operation Context to the ReadTpr implementation."""
    # TODO: Either get rid of **kwargs or clarify the roadmap and timeline for doing so.
    filename = input.kwargs["_simulation_input"]
    parameters = input.kwargs["parameters"]
    return SessionResources(
        _simulation_input=filename, parameters=parameters, publisher=output
    )


def _standard_node_resource_factory(*args, **kwargs):
    """Translate Python UI input to the gmxapi.operation node builder inputs."""
    return _input.bind(*args, **kwargs)


def _run(resources: SessionResources):
    """Operation implementation in the gmxapi.operation module context."""
    # We combine the input from *input*, and key word arguments into the new output.
    for named_data in resources.output.keys():
        assert isinstance(named_data, str)
        assert hasattr(resources, named_data)
        source_value = getattr(resources, named_data)
        logger.debug(
            "modify_input publishing {} to {}".format(source_value, named_data)
        )
        setattr(resources.output, named_data, source_value)


# Note: this is a class because we attach the input_description functionality,
# but all of its functionality can be composed from dispatching functions.
# TODO: Consider replacing this unique class with a pattern of composed instances of a common class.
class ResourceFactory(gmxapi.abc.ResourceFactory):
    """ModifyInput resource factory."""

    def __init__(self, target_context, source_context):
        self.source_context = source_context  # Determine input form of *create* method.
        self.target_context = target_context  # Context in which resources are consumed.

    # TODO: clean up the dispatching. What to do when input comes from multiple sources?
    # Use a typing overload or a single-dispatch functor for clearer input/result typing.
    def __call__(self, *args, **kwargs):
        # context is used for dispatching and annotates the Context of the other arguments.
        # context == None implies inputs are from Python UI.
        if self.source_context is None:
            # TODO: De-duplicate re: StandardDirector.resource_factory()
            if isinstance(self.target_context, _op.Context):
                return _standard_node_resource_factory(*args, **kwargs)
        if isinstance(self.source_context, _op.Context):
            # TODO: Check whether the consumer is a Context.NodeBuilder or an operation runner.
            # We don't yet use this dispatcher for building nodes, so assume we are launching a session.
            assert "input" in kwargs
            assert "output" in kwargs
            return _session_resource_factory(
                input=kwargs["input"], output=kwargs["output"]
            )
        raise gmxapi.exceptions.MissingImplementationError(
            "No translation from {} context to {}".format(
                self.source_context, self.target_context
            )
        )

    @typing.overload
    def input_description(self, context: _op.Context) -> _op.InputDescription: ...

    def input_description(self, context: gmxapi.abc.Context):
        if isinstance(context, _op.Context):
            return StandardInputDescription()
        raise gmxapi.exceptions.MissingImplementationError(
            "No input description available for {} context".format(context)
        )


class StandardInputDescription(_op.InputDescription):
    """Provide the ModifyInput input description in gmxapi.operation Contexts."""

    # TODO: Improve fingerprinting.
    # If _make_uid can't make sufficiently unique IDs, use additional "salt".
    # Without fingerprinting, we cannot consistently hash *input* across processes,
    # but we can consistently generate integers in the same sequence the first time
    # we see each distinct input.
    _next_uid: typing.ClassVar[int] = 0
    _uids: typing.ClassVar[typing.MutableMapping[int, int]] = {}

    @classmethod
    def _make_uid(cls, input) -> str:
        # TODO: Use input fingerprint for more useful identification.
        # WARNING: The built-in hash will use memory locations, and so will not be consistent across
        # process ranks, even if the input should be the same.
        salt = hash(input)
        if salt not in cls._uids:
            cls._uids[salt] = cls._next_uid
            cls._next_uid += 1
        else:
            logger.debug(f"Reissuing uid for modify_input({input}): {cls._uids[salt]}")
        new_uid = "modify_input_{}".format(cls._uids[salt])
        return new_uid

    def signature(self) -> InputCollectionDescription:
        return _input

    def make_uid(self, input: _op.DataEdge) -> str:
        assert isinstance(input, _op.DataEdge)
        return self._make_uid(input)


class RegisteredOperation(_op.OperationImplementation, metaclass=_op.OperationMeta):
    """Provide the gmxapi compatible ModifyInput implementation."""

    # This is a class method to allow the class object to be used in gmxapi.operation._make_registry_key
    @classmethod
    def name(self) -> str:
        """Canonical name for the operation."""
        return "modify_input"

    @classmethod
    def namespace(self) -> str:
        """modify_input is importable from the gmxapi module."""
        return "gmxapi"

    @classmethod
    def director(cls, context: gmxapi.abc.Context):
        # Currently, we only have a Directory for the gmxapi.operation.Context
        if isinstance(context, _op.Context):
            return StandardDirector(context)


class StandardOperationHandle(_op.AbstractOperation, ModuleObject):
    """Handle used in Python UI or gmxapi.operation Contexts."""

    def __init__(self, resource_manager: ResourceManager):
        self.__resource_manager = resource_manager

    def run(self):
        self.__resource_manager.update_output()

    @property
    def output(self) -> OutputDataProxy:
        return self.__resource_manager.data()


class StandardDirector(gmxapi.abc.OperationDirector):
    """Direct the instantiation of a modify_input node in a gmxapi.operation Context.

    .. todo:: Compose this behavior in a more generic class.

    .. todo:: Describe where instances live.
    """

    def __init__(self, context: _op.Context):
        if not isinstance(context, _op.Context):
            raise gmxapi.exceptions.ValueError(
                "StandardDirector requires a gmxapi.operation Context."
            )
        self.context = context

    def __call__(
        self, resources: _op.DataSourceCollection, label: str = None
    ) -> StandardOperationHandle:
        builder = self.context.node_builder(operation=RegisteredOperation, label=label)

        builder.set_resource_factory(_session_resource_factory)
        builder.set_input_description(StandardInputDescription())
        builder.set_handle(StandardOperationHandle)

        runner_director = _op.RunnerDirector(runner=_run, allow_duplicate=True)

        builder.set_runner_director(runner_director)
        builder.set_output_factory(_output_factory)
        builder.set_resource_manager(ResourceManager)

        # Note: we have not yet done any dispatching based on *resources*. We should
        # translate the resources provided into the form that the Context can subscribe to
        # using the dispatching resource_factory.
        assert isinstance(resources, _op.DataSourceCollection)
        for name, source in resources.items():
            builder.add_input(name, source)

        handle = builder.build()
        assert isinstance(handle, StandardOperationHandle)
        return handle

    def handle_type(self, context: gmxapi.abc.Context):
        return StandardOperationHandle

    def resource_factory(
        self,
        source: typing.Union[gmxapi.abc.Context, ModuleObject, None],
        target: gmxapi.abc.Context = None,
    ):
        # Distinguish between the UIContext, in which input is in the form
        # of function call arguments, and the StandardContext, implemented in
        # gmxapi.operation. UIContext is probably a virtual context that is
        # asserted by callers in order to get a factory that normalizes UI input
        # for the StandardContext.
        #
        if target is None:
            target = self.context
        # TODO: Normalize and consolidate the context-based dispatching.
        if source is None:
            # `source is None` implies source is coming from UI.
            if isinstance(target, _op.Context):
                # Return a factory that can bind to function call arguments to produce a DataSourceCollection.
                return ResourceFactory(target_context=target, source_context=source)
        if isinstance(source, _op.Context):
            # The source is a gmxapi.operation.Context when the operation is being evaluated through
            # a gmxapi.operation.ResourceManager. i.e. at run time.
            return ResourceFactory(target_context=target, source_context=source)
        if isinstance(source, ModuleObject):
            if isinstance(target, _op.Context):
                # We are creating a node in gmxapi.operation.Context from another gmxapi.simulation operation.
                # This means that we want to subscribe to the subcontext instead of the gmxapi.operation.Context.
                # In the first draft, though, we just access a special payload.
                # Return a factory that will consume *_simulation_input* and *parameters*
                # members of a received object.
                logger.info("Building mdrun operation from source {}".format(source))

                def simulation_input_workaround(input, parameters: dict):
                    source = input
                    # TODO: Normalize mechanism for obtaining SimulationInput references.
                    if hasattr(source, "output"):
                        source = input.output
                    assert hasattr(source, "_simulation_input")
                    assert hasattr(source, "parameters")
                    logger.info(
                        "modify_input receiving input {}: {}".format(
                            source._simulation_input.name,
                            source._simulation_input.description,
                        )
                    )
                    source_collection = _input.bind(
                        _simulation_input=source._simulation_input,
                        parameters=parameters,
                    )
                    logger.info(
                        "modify_input input bound as source collection {}".format(
                            source_collection
                        )
                    )
                    return source_collection

                return simulation_input_workaround

        raise gmxapi.exceptions.ValueError(
            "No dispatching from {} context to {}".format(source, target)
        )


# TODO: This operation is intended to be able to compose complete simulation input. E.g:
#  def modify_input(input: SimulationInput, parameters=None, topology=None, conformation=None, simulation_state=None):
def modify_input(input, parameters: dict, label: str = None, context=None):
    """Modify simulation input with data flow operations.

    Given simulation input *input*, override components of simulation input with
    additional arguments, such as *parameters*.

    See :py:class:`~gmxapi.simulation.modify_input.OutputDataProxy` for *output*
    attribute members.
    """
    handle_context = context
    if handle_context is not None:
        raise gmxapi.exceptions.MissingImplementationError(
            "context must be None. This factory is only for the Python UI right now."
        )

    target_context = _op.current_context()
    assert isinstance(target_context, _op.Context)
    # Get a director that will create a node in the standard context.
    # TODO: For clarity, restructure code to use module-local helper functors that produce callables with clear
    #  interfaces. (May clash with the current ABC scheme.)
    node_director: StandardDirector = _op._get_operation_director(
        RegisteredOperation, context=target_context
    )
    assert isinstance(node_director, StandardDirector)
    # TODO: refine this protocol
    assert handle_context is None
    # Examine the input.
    if isinstance(input, ModuleObject):
        # The input is from read_tpr or modify_input.
        source_context = input
    else:
        # Allow automatic dispatching
        source_context = None

    # TODO: Fix this protocol to do dispatching in the correct place.
    # The source Context here is None (the handle Context). The resources themselves
    # may be from different Contexts, so we should dispatch at the add_resource
    # builder method, not here in the director client.
    resource_factory: ResourceFactory = node_director.resource_factory(
        source=source_context, target=target_context
    )
    resources: _op.DataSourceCollection = resource_factory(
        input=input, parameters=parameters
    )
    handle = node_director(resources=resources, label=label)
    return handle



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2014- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Doxygen XML output parser.

This module implements a parser for the Doxygen XML output, converting it into
an object model that can be used to navigate the documentation.  It also uses
knowledge from how Doxygen works to provide access to things like visibility of
individual member documentation (e.g., based on what is the visibility of its
parent compound objects).

The object model is rooted at a DocumentationSet object.  Each documented
entity is modeled as an Entity, and this has subclasses Member and Compound to
correspond to the two categories of items that Doxygen handles.  These classes
are further subclassed to match each kind of entity that Doxygen produces.
Only kinds produced by Doxygen from C/C++ code are modeled.  Everything else
is ignored after a warning.

Currently the member entities are not completely parsed from the XML files, and
the interface may need additional work to provide convenient access to all
member types and their common properties.  For now, focus is in modeling the
compound entities.

The implementation is mostly independent of any GROMACS-specific rules, except
for the following:
 - DocType.library is a GROMACS-specific construct that is deduced from the
   contents of the detailed description (presence of a \libinternal command in
   the Doxygen comment triggers it).
 - DocType.internal is deduced from the presence of a \internal command that
   covers the whole detailed description.
 - List of extensions for determining whether a file is a source file only
   contains extensions actually used by GROMACS.
It would be possible to move these out from this file, but that would require
exposing the XML representation for the descriptions, which is not nice either.

The module can also be run as a script that can dump out different parts of the
object model.  This can be used to debug the parser, as well as check what is
actually in the XML documentation.
"""

import os.path
import xml.etree.ElementTree as ET
import functools

import reporter

#####################################################################
# Helper functions and classes


def _show_list(title, objlist):
    """Helper function for formatting a list of objects for debug output."""
    if objlist:
        print("{0}:".format(title))
        for obj in objlist:
            print("  ", obj)


@functools.total_ordering
class DocType(object):
    """Documentation visibility in the generated documentation."""

    # Mapping to string representations for the internal integer values
    _names = ["undocumented", "internal", "library", "public"]

    def __init__(self, value):
        """Initialize a DocType instance.

        DocType.{none,internal,library,public} should be used outside the class
        instead of calling the constructor.
        """
        self._value = value

    def __str__(self):
        """Return string representation for the documentation type."""
        return self._names[self._value]

    def __eq__(self, other):
        """Order documentation types in the order of visibility."""
        return self._value == other._value

    def __lt__(self, other):
        """Order documentation types in the order of visibility."""
        return self._value < other._value


# Static values for documentation types.
DocType.none = DocType(0)
DocType.internal = DocType(1)
DocType.library = DocType(2)
DocType.public = DocType(3)


class Location(object):
    """Location of a Doxygen entity.

    This class contains the logic to parse a <location> tag in Doxygen XML.
    It is used as the entity location in cases where body location is not
    expected, or as part of a LocationWithBody.
    """

    def __init__(self, elem):
        """Initialize location from a <location> element."""
        self.filepath = elem.attrib["file"]
        self.line = int(elem.attrib["line"])
        self.column = elem.attrib["column"]

    def __str__(self):
        return "{0}:{1}".format(self.filepath, self.line)

    def get_reporter_location(self):
        return reporter.Location(self.filepath, self.line)

    def get_full_string(self):
        return "{0}:{1}:{2}".format(self.filepath, self.line, self.column)


@functools.total_ordering
class BodyLocation(object):
    """Body location of a Doxygen entity.

    This class contains the logic to parse a body location from a <location>
    tag in Doxygen XML.  Not all entities have these attributes.
    This is only used as part of a LocationWithBody, which handles cases where
    the body location is optional.

    The body location can be compared and hashed so that it can be used in
    a dictionary for DocumentationSet.merge_duplicates().
    """

    def __init__(self, elem):
        """Initialize body location from a <location> element."""
        self.filepath = elem.attrib["bodyfile"]
        self.startline = int(elem.attrib["bodystart"])
        self.endline = int(elem.attrib["bodyend"])

    def __eq__(self, other):
        return (
            self.filepath == other.filepath
            and self.startline == other.startline
            and self.endline == other.endline
        )

    def __lt__(self, other):
        if self.filepath != other.filepath:
            if other.filepath is None:
                return False
            if self.filepath is None:
                return True
            return self.filepath < other.filepath
        elif self.startline != other.startline:
            return self.startline < other.startline
        else:  # check for endline
            return self.endline < other.endline

    def __hash__(self):
        return hash(self.filepath) ^ hash(self.startline) ^ hash(self.endline)

    def __str__(self):
        return "{0}:{1}".format(self.filepath, self.startline)

    def get_full_string(self):
        if self.endline < 0:
            return self.__str__()
        return "{0}:{1}-{2}".format(self.filepath, self.startline, self.endline)


class LocationWithBody(object):
    """Location for a Doxygen entity that can have a body location.

    This class is used to represent the location of a Doxygen entity that can
    have a body location.
    """

    def __init__(self, elem):
        """Initialize location from a <location> element."""
        self._location = Location(elem)
        if "bodyfile" in elem.attrib:
            self._bodylocation = BodyLocation(elem)
        else:
            self._bodylocation = None

    def __str__(self):
        if not self._bodylocation:
            return "{0} (no body)".format(self._location)
        else:
            return "{0} / {1}".format(self._location, self._bodylocation)

    def get_reporter_location(self):
        """Return reporter location for this location.

        All issues are reported at the main location, which should match with
        the declaration, where most of the documentation typically is.
        """
        return self._location.get_reporter_location()

    def get_location(self):
        return self._location

    def get_body_location(self):
        return self._bodylocation

    def has_same_body_location(self):
        """Check whether main location matches body location.

        If the main location is different, then it likely points to the
        declaration of the function.
        """
        return (
            self._location.filepath == self._bodylocation.filepath
            and self._location.line == self._bodylocation.startline
        )


class MemberSection(object):
    """Section of members within a compound entity."""

    def __init__(self, kind):
        self._kind = kind
        self._members = []

    def __str__(self):
        return self._kind

    def add_member(self, member):
        self._members.append(member)

    def replace_member(self, old, new):
        try:
            pos = self._members.index(old)
        except ValueError:
            return
        self._members[pos] = new


#####################################################################
# Documentation entities


class Entity(object):
    """Doxygen documentation entity.

    This class represents common properties of an entity that can contain
    Doxygen documentation.
    """

    def __init__(self, name, refid):
        self._docset = None
        self._name = name
        self._id = refid
        self._has_brief_description = False
        self._has_detailed_description = False
        self._has_inbody_description = False
        self._visibility = DocType.none

    def __str__(self):
        return self._name

    def _get_reporter(self):
        """Return reporter to use for parsing issues."""
        return self._docset.get_reporter()

    def set_documentation_set(self, docset):
        """Set the documentation set this entity belongs to.

        The documentation set parent provides access to a common reporter
        object, and also allows the entity to resolve references to other
        entities while loading XML information.
        """
        assert self._docset is None
        self._docset = docset

    def get_id(self):
        return self._id

    def get_name(self):
        return self._name

    def get_reporter_location(self):
        return reporter.Location("<{0}>".format(self._name), None)

    def get_visibility(self):
        return self._visibility

    def is_documented(self):
        return self._visibility != DocType.none

    def has_brief_description(self):
        return self._has_brief_description

    def has_inbody_description(self):
        return self._has_inbody_description

    def _process_descriptions(self, briefelem, detailselem, inbodyelem):
        reporter = self._get_reporter()
        if briefelem is not None and len(briefelem) > 0:
            self._has_brief_description = True
            self._visibility = DocType.public
        if detailselem is not None and len(detailselem) > 0:
            self._visibility = DocType.public
            # Gromacs-specific:
            # \internal is used at the beginning of a comment block to
            # mark the block internal to the module.
            # \libinternal is used similarly, and inserts custom XML
            # elements.
            if detailselem[0].tag == "internal":
                if len(detailselem) == 1:
                    self._visibility = DocType.internal
                else:
                    # TODO: Should we also check if internal appears elsewhere?
                    reporter.doc_note(
                        self, "\internal does not cover whole documentation"
                    )
            if detailselem[0].find("libinternal") is not None:
                if self._visibility == DocType.public:
                    self._visibility = DocType.library
                else:
                    reporter.doc_error(
                        self, "\libinternal should not be used inside \internal"
                    )
            self._has_detailed_description = True
        if inbodyelem is not None:
            self._has_inbody_description = len(inbodyelem) > 0

    def show_base(self):
        """Format information for common properties.

        This is called from subclass show() methods to show base information
        about the entity.
        """
        print("ID:         {0}".format(self._id))
        print("Name:       {0}".format(self._name))
        print("Location:   {0}".format(self.get_reporter_location()))
        doctype = []
        if self._has_brief_description:
            doctype.append("brief")
        if self._has_detailed_description:
            doctype.append("details")
        if self._has_inbody_description:
            doctype.append("in-body")
        if not doctype:
            doctype.append("none")
        print("Doc:        {0}".format(", ".join(doctype)))
        print("Visibility: {0}".format(self._visibility))


# Member entities


class Member(Entity):
    """Member entity.

    In Doxygen, a member entity is an entity such as a function or an enum that
    cannot contain other documented entities (an enum is a slight exception, as
    enum values are still nested within the enum member).  A member always
    belongs to one (or more) compounds, which means that the detailed
    documentation for the member appears on the documentation page for that
    compound.  If none of the parent compounds are documented, the member
    doesn't appear anywhere, even if it is documented.

    Member information is loaded from a parent compound's XML file.  If there
    is more than one parent, the first one encountered will be used
    (presumably, Doxygen duplicates the information into each XML file).
    """

    def __init__(self, name, refid):
        Entity.__init__(self, name, refid)
        self._parents = set()
        self._class = None
        self._namespace = None
        self._files = set()
        self._group = None
        self._location = None
        self._alternates = set()
        self._loaded = False
        # TODO: Move to Entity?
        self._xmlpath = None

    def add_parent_compound(self, compound):
        """Add a compound that contains this member."""
        self._parents.add(compound)
        if isinstance(compound, Class):
            assert (
                self._class is None
            ), 'Class "{0}" was already added. Maybe you have two entities with the same name {1}'.format(
                self._class, self._name
            )
            self._class = compound
        elif isinstance(compound, Namespace):
            assert (
                self._namespace is None
            ), 'Namespace "{0}" was already added. Maybe you have two entities with the same name {1}'.format(
                self._namespace, self._name
            )
            self._namespace = compound
        elif isinstance(compound, File):
            self._files.add(compound)
        elif isinstance(compound, Group):
            assert self._group is None, 'Group "{0}" was already added.'.format(
                self._group
            )
            self._group = compound
        else:
            assert False

    def merge_definition(self, definition):
        """Merge another member into this.

        See DocumentationSet.merge_duplicates().
        """
        assert self._class is None
        assert definition._class is None
        assert self._group == definition._group
        assert self._namespace == definition._namespace
        self._parents.update(definition._parents)
        self._files.update(definition._files)
        self._alternates.add(definition)

    def load_details_from_element(self, rootelem, xmlpath):
        """Load details for the member from a given XML element.

        This method is called when encountering member definitions while
        processing a compound XML file to load the information for that member.
        It processes common properties for a member, and delegates other
        elements to _load_element().
        """
        if self._loaded:
            # TODO: It would be nice to verify that the same information
            # is present in all instances
            return
        self._xmlpath = xmlpath
        # TODO: Process the attributes
        reporter = self._get_reporter()
        briefelem = None
        detailselem = None
        inbodyelem = None
        for elem in rootelem:
            if elem.tag == "name":
                if elem.text != self.get_name():
                    reporter.xml_assert(
                        xmlpath,
                        "member name mismatch: '{0}' (in index.xml) vs. '{1}'".format(
                            self.get_name(), elem.text
                        ),
                    )
            elif elem.tag == "briefdescription":
                briefelem = elem
            elif elem.tag == "detaileddescription":
                detailselem = elem
            elif elem.tag == "inbodydescription":
                # TODO: in-body description is probably only possible for
                # functions; move it there.
                inbodyelem = elem
            elif elem.tag == "location":
                self._location = LocationWithBody(elem)
            else:
                if not self._load_element(elem):
                    # TODO Process the rest of the elements so that we can check this
                    # reporter.xml_assert(xmlpath,
                    #        "unknown member child element '{0}'".format(elem.tag))
                    pass
        self._process_descriptions(briefelem, detailselem, inbodyelem)
        self._loaded = True

    def _load_element(self, element):
        """Load data from a child XML element.

        This method is called for all XML elements under the <memberdef>
        element that are not handled directly by the Member class.
        Derived classes should return True if they process the element.
        """
        return False

    def _get_raw_location(self):
        """Returns the BodyLocation object associated with this member.

        This is necessary so that EnumValue can override it report a non-empty
        location: Doxygen doesn't provide any location for <enumvalue>.
        """
        return self._location

    def get_reporter_location(self):
        return self._get_raw_location().get_reporter_location()

    def get_location(self):
        """Return main location for the member.

        This typically corresponds to the declaration.
        """
        return self._get_raw_location().get_location()

    def get_body_location(self):
        """Return location of the body for the member.

        Some types of members do not have a body location, in which case this
        returns None.
        """
        return self._get_raw_location().get_body_location()

    def has_same_body_location(self):
        """Check whether the main location is the same as body location."""
        return self._get_raw_location().has_same_body_location()

    def get_namespace(self):
        return self._namespace

    def get_parent_compounds(self):
        return self._parents

    def get_inherited_visibility(self):
        return max([parent.get_visibility() for parent in self._parents])

    def show(self):
        self.show_base()
        if self._alternates:
            idlist = [x.get_id() for x in self._alternates]
            print("Alt. IDs:   {0}".format(", ".join(idlist)))
        print("Parent vis: {0}".format(self.get_inherited_visibility()))
        print("Location:   {0}".format(self.get_location().get_full_string()))
        print("Body loc:   {0}".format(self.get_body_location().get_full_string()))
        _show_list("Parents", self._parents)


class Define(Member):
    pass


class Variable(Member):
    pass


class Typedef(Member):
    pass


class Enum(Member):
    def __init__(self, name, refid):
        Member.__init__(self, name, refid)
        self._values = set()

    def _load_element(self, elem):
        if elem.tag == "enumvalue":
            refid = elem.attrib["id"]
            # Doxygen seems to sometimes assign the same ID to a singleton enum
            # value (this already triggers a warning in loading index.xml).
            if refid == self.get_id():
                return True
            member = self._docset.get_member(refid)
            member.set_enum(self)
            member.load_details_from_element(elem, self._xmlpath)
            self._values.add(member)
            return True
        return False

    def get_values(self):
        return self._values


class EnumValue(Member):
    def __init__(self, name, refid):
        Member.__init__(self, name, refid)
        self._enum = None

    def set_enum(self, member):
        assert self._enum is None
        self._enum = member

    def _get_raw_location(self):
        return self._enum._get_raw_location()


class Function(Member):
    pass


class FriendDeclaration(Member):
    pass


# Compound entities


class Compound(Entity):
    """Compound entity.

    In Doxygen, a compound entity is an entity that has its own documentation
    page, and can contain other documented entities (either members, or other
    compounds).  Examples of compounds are files and classes.
    A compound entity always appears in the documentation, even if it is
    contained in another compound that is not documented.

    The list of members for a compound is initialized when the XML index file
    is read.  All other information is loaded from an XML file that is specific
    to the compound.  In addition to describing the compound, this XML file
    contains references to contained compounds, and details of all members
    within the compound.
    """

    def __init__(self, name, refid):
        Entity.__init__(self, name, refid)
        self._members = dict()
        self._children = set()
        self._sections = []
        self._groups = set()
        self._loaded = False

    def get_xml_path(self):
        """Return path to the details XML file for this compound."""
        return os.path.join(self._docset.get_xmlroot(), self.get_id() + ".xml")

    def add_member(self, member):
        """Add a contained member."""
        self._members[member.get_id()] = member

    def add_group(self, compound):
        """Add a group (a compound entity) that contains this entity."""
        self._groups.add(compound)

    def replace_member(self, old, new):
        if old.get_id() not in self._members:
            raise ValueError("Trying to replace a non-existent member")
        elif new.get_id() in self._members:
            raise ValueError("Trying to replace with an existing member")
        self._members[old.get_id()] = new
        for section in self._sections:
            section.replace_member(old, new)

    def load_details(self):
        """Load details for the compound from its details XML file.

        This method processes common properties for a compound.
        References to inner compounds are delegated to _load_inner_*() methods,
        and all members encountered in the XML file are loaded with
        Member.load_details_from_element().
        Other elements are delegated to _load_element().
        """
        if self._loaded:
            return
        reporter = self._get_reporter()
        xmlpath = self.get_xml_path()
        compoundtree = ET.parse(xmlpath)
        root = compoundtree.getroot()
        if len(root) > 1:
            reporter.xml_assert(xmlpath, "more than one compound in a file")
        if root[0].tag != "compounddef":
            reporter.xml_assert(xmlpath, "expected <compounddef> as the first tag")
            return
        briefelem = None
        detailselem = None
        missing_members = set(self._members.values())
        for elem in root[0]:
            if elem.tag == "compoundname":
                if elem.text != self.get_name():
                    reporter.xml_assert(
                        xmlpath,
                        "compound name mismatch: '{0}' (in index.xml) vs. '{1}'".format(
                            self.get_name(), elem.text
                        ),
                    )
            elif elem.tag == "briefdescription":
                briefelem = elem
            elif elem.tag == "detaileddescription":
                detailselem = elem
            elif elem.tag in (
                "includes",
                "includedby",
                "incdepgraph",
                "invincdepgraph",
                "inheritancegraph",
                "collaborationgraph",
                "programlisting",
                "templateparamlist",
                "listofallmembers",
            ):
                pass
            elif elem.tag.startswith("inner"):
                refid = elem.attrib["refid"]
                reftype = elem.tag[5:]
                # TODO: Handle 'prot' attribute?
                refcompound = self._docset.get_compound(refid)
                self._children.add(refcompound)
                if reftype == "file":
                    self._load_inner_file(refcompound)
                elif reftype == "dir":
                    self._load_inner_dir(refcompound)
                elif reftype == "group":
                    self._load_inner_group(refcompound)
                elif reftype == "namespace":
                    self._load_inner_namespace(refcompound)
                elif reftype == "class":
                    self._load_inner_class(refcompound)
                else:
                    reporter.xml_assert(
                        xmlpath, "unknown inner compound type '{0}'".format(reftype)
                    )
            elif elem.tag == "sectiondef":
                # TODO: Handle header and description elements
                kind = elem.attrib["kind"]
                section = MemberSection(kind)
                self._sections.append(section)
                for memberelem in elem.iter("memberdef"):
                    refid = memberelem.attrib["id"]
                    member = self._members[refid]
                    member.load_details_from_element(memberelem, xmlpath)
                    section.add_member(member)
                    if member in missing_members:
                        missing_members.remove(member)
                    # Enum values need special handling, but are not worth
                    # extra generalization.
                    if isinstance(member, Enum):
                        missing_members.difference_update(member.get_values())
            else:
                if not self._load_element(elem):
                    reporter.xml_assert(
                        xmlpath, "unknown compound child element '{0}'".format(elem.tag)
                    )
        if missing_members:
            reporter.xml_assert(xmlpath, "members without section")
        self._process_descriptions(briefelem, detailselem, None)
        self._loaded = True

    def _unexpected_inner_compound(self, typename, compound):
        """Report a parsing error for an unexpected inner compound reference."""
        reporter = self._get_reporter()
        xmlpath = self.get_xml_path()
        reporter.xml_assert(
            xmlpath, "unexpected inner {0}: {1}".format(typename, compound)
        )

    def _load_inner_file(self, compound):
        """Process a reference to an inner file.

        Derived classes should override the method if the compound type can
        contain files as nested compounds.
        """
        self._unexpected_inner_compound("file", compound)

    def _load_inner_dir(self, compound):
        """Process a reference to an inner directory.

        Derived classes should override the method if the compound type can
        contain directories as nested compounds.
        """
        self._unexpected_inner_compound("dir", compound)

    def _load_inner_group(self, compound):
        """Process a reference to an inner group.

        Derived classes should override the method if the compound type can
        contain groups as nested compounds.
        """
        self._unexpected_inner_compound("group", compound)

    def _load_inner_namespace(self, compound):
        """Process a reference to an inner namespace.

        Derived classes should override the method if the compound type can
        contain namespaces as nested compounds.
        """
        self._unexpected_inner_compound("namespace", compound)

    def _load_inner_class(self, compound):
        """Process a reference to an inner class.

        Derived classes should override the method if the compound type can
        contain classes as nested compounds.
        """
        self._unexpected_inner_compound("class", compound)

    def _load_element(self, element):
        """Load data from a child XML element.

        This method is called for all XML elements under the <compounddef>
        element that are not handled directly by the Compound class.
        Derived classes should return True if they process the element.
        """
        return False

    def get_groups(self):
        return self._groups

    def show_base(self):
        """Format information for common properties.

        This extends Entity.show_base() by adding properties that are common to
        all compounds.
        """
        Entity.show_base(self)
        if self._groups:
            print("Groups:   {0}".format(", ".join(map(str, self._groups))))

    def show_members(self):
        """Show list of members.

        This method is provided for use in show() methods of derived classes
        to print the list of members.
        """
        for section in self._sections:
            print("Member section: {0}".format(section))
            for member in section._members:
                print("  ", member)


class File(Compound):
    def __init__(self, name, refid):
        Compound.__init__(self, name, refid)
        self._path = None
        self._directory = None
        self._classes = set()
        self._namespaces = set()
        self._is_source_file = None

    def _load_inner_class(self, compound):
        compound.add_file(self)
        self._classes.add(compound)

    def _load_inner_namespace(self, compound):
        compound.add_file(self)
        self._namespaces.add(compound)

    def _load_element(self, elem):
        if elem.tag == "location":
            self._path = elem.attrib["file"]
            extension = os.path.splitext(self._path)[1]
            self._is_source_file = extension in (".c", ".cpp", ".cu")
            return True
        return False

    def set_directory(self, directory):
        self._directory = directory

    def get_reporter_location(self):
        return reporter.Location(self._path, None)

    def get_path(self):
        return self._path

    def get_directory(self):
        return self._directory

    def is_source_file(self):
        return self._is_source_file

    def show(self):
        self.show_base()
        print("Path:      {0}".format(self._path))
        print("Directory: {0}".format(self._directory))
        print("Source:    {0}".format(self._is_source_file))
        _show_list("Namespaces", self._namespaces)
        _show_list("Classes", self._classes)
        self.show_members()


class Directory(Compound):
    def __init__(self, name, refid):
        Compound.__init__(self, name, refid)
        self._path = None
        self._parent = None
        self._subdirs = set()
        self._files = set()

    def _load_inner_file(self, compound):
        compound.set_directory(self)
        self._files.add(compound)

    def _load_inner_dir(self, compound):
        compound._parent = self
        self._subdirs.add(compound)

    def _load_element(self, elem):
        if elem.tag == "location":
            self._path = elem.attrib["file"]
            return True
        return False

    def get_reporter_location(self):
        return reporter.Location(self._path, None)

    def get_path(self):
        return self._path

    def get_parent(self):
        return self._parent

    def get_subdirectories(self):
        return self._subdirs

    def show(self):
        self.show_base()
        print("Path:      {0}".format(self._path))
        if self._parent:
            print("Parent:    {0}".format(self._parent))
        _show_list("Subdirectories", self._subdirs)
        _show_list("Files", self._files)


class Group(Compound):
    def __init__(self, name, refid):
        Compound.__init__(self, name, refid)
        self._title = None
        self._files = set()
        self._nestedgroups = set()
        self._namespaces = set()
        self._classes = set()

    def _load_inner_file(self, compound):
        compound.add_group(self)
        self._files.add(compound)

    # Doxygen 1.8.5 doesn't seem to put the directories into the XML output,
    # even though they are in the HTML output as group members...

    def _load_inner_group(self, compound):
        compound.add_group(self)
        self._nestedgroups.add(compound)

    def _load_inner_namespace(self, compound):
        compound.add_group(self)
        self._namespaces.add(compound)

    def _load_inner_class(self, compound):
        compound.add_group(self)
        self._classes.add(compound)

    def _load_element(self, elem):
        if elem.tag == "title":
            self._title = elem.text
            return True
        return False

    def show(self):
        self.show_base()
        print("Title:     {0}".format(self._title))
        print("Inner compounds:")
        for compound in self._children:
            print("  ", compound)
        self.show_members()


class Namespace(Compound):
    def __init__(self, name, refid):
        Compound.__init__(self, name, refid)
        self._doclocation = None
        self._files = set()
        self._parent = None
        self._innernamespaces = set()
        self._classes = set()

    def _load_inner_namespace(self, compound):
        compound._parent = self
        self._innernamespaces.add(compound)

    def _load_inner_class(self, compound):
        compound.set_namespace(self)
        self._classes.add(compound)

    def _load_element(self, elem):
        if elem.tag == "location":
            self._doclocation = Location(elem)
            return True
        return False

    def add_file(self, compound):
        self._files.add(compound)

    def get_reporter_location(self):
        return self._doclocation.get_reporter_location()

    def is_anonymous(self):
        return "anonymous_namespace{" in self.get_name()

    def show(self):
        self.show_base()
        print("Doc. loc.: {0}".format(self._doclocation.get_full_string()))
        _show_list("Inner namespaces", self._innernamespaces)
        _show_list("Classes", self._classes)
        self.show_members()


class Class(Compound):
    def __init__(self, name, refid):
        Compound.__init__(self, name, refid)
        self._location = None
        self._namespace = None
        self._files = set()
        self._baseclasses = []
        self._derivedclasses = set()
        self._outerclass = None
        self._innerclasses = set()

    def _load_inner_class(self, compound):
        compound.set_outer_class(self)
        self._innerclasses.add(compound)

    def _load_element(self, elem):
        if elem.tag == "basecompoundref":
            # TODO: Handle unknown bases?
            if "refid" in elem.attrib:
                refid = elem.attrib["refid"]
                # TODO: Handle prot and virt attributes, check name?
                base = self._docset.get_compound(refid)
                self._baseclasses.append(base)
            return True
        if elem.tag == "derivedcompoundref":
            refid = elem.attrib["refid"]
            # TODO: Handle prot and virt attributes, check name?
            derived = self._docset.get_compound(refid)
            self._derivedclasses.add(derived)
            return True
        elif elem.tag == "location":
            self._location = LocationWithBody(elem)
            return True
        return False

    def add_file(self, compound):
        self._files.add(compound)

    def set_namespace(self, compound):
        self._namespace = compound

    def set_outer_class(self, compound):
        self._outerclass = compound

    def get_reporter_location(self):
        return self._location.get_reporter_location()

    def get_files(self):
        return self._files

    def is_local(self):
        if len(self._files) > 1:
            return False
        for fileobj in self._files:
            if not fileobj.is_source_file():
                return False
        return True

    def show(self):
        self.show_base()
        print("Namespace:  {0}".format(self._namespace))
        if self._outerclass:
            print("Outer cls:  {0}".format(self._outerclass))
        location = self._location
        print("Location:   {0}".format(location.get_location().get_full_string()))
        print("Body loc:   {0}".format(location.get_body_location().get_full_string()))
        _show_list("Inner classes", self._innerclasses)
        self.show_members()


#####################################################################
# Top-level container class


def _get_compound_type_from_kind(kind):
    """Map compound kinds from Doxygen XML to internal class types."""
    if kind == "file":
        return File
    elif kind == "dir":
        return Directory
    elif kind == "group":
        return Group
    elif kind == "namespace":
        return Namespace
    elif kind in ("class", "struct", "union"):
        return Class
    else:
        return None


def _get_member_type_from_kind(kind):
    """Map member kinds from Doxygen XML to internal class types."""
    if kind == "define":
        return Define
    elif kind == "variable":
        return Variable
    elif kind == "typedef":
        return Typedef
    elif kind == "enum":
        return Enum
    elif kind == "enumvalue":
        return EnumValue
    elif kind == "function":
        return Function
    elif kind == "friend":
        return FriendDeclaration
    else:
        return None


class DocumentationSet(object):
    """Root object for Doxygen XML documentation tree.

    On initialization, it reads the index.xml file from the Doxygen XML output,
    which contains the list of entities.  Only the ID and name for the entities,
    and the parent compounds for members, are available from this file.

    load_details() can be called to load the detailed compound XML files.
    This constructs relations between compound entities, and initializes other
    attributes for the entities.

    load_file_details() does the same as load_details(), except that it leaves
    those compound XML files unloaded that do not affect file objects or their
    parent hierarchy.  This saves some time if details for actual code
    constructs like namespaces, classes or members are not necessary.

    merge_duplicates() can then be called to remove members with different IDs,
    but that actually reference the same code entity.  For some reason, Doxygen
    seems to produce these in certain cases.
    """

    def __init__(self, xmlroot, reporter):
        """Initialize the documentation set and read index data."""
        self._xmlroot = xmlroot
        self._reporter = reporter
        xmlpath = os.path.join(xmlroot, "index.xml")
        indextree = ET.parse(xmlpath)
        self._compounds = dict()
        self._members = dict()
        self._files = dict()
        for compoundelem in indextree.getroot():
            name = compoundelem.find("name").text
            refid = compoundelem.attrib["refid"]
            kind = compoundelem.attrib["kind"]
            if kind in ("page", "example"):
                # TODO: Model these types as well
                continue
            compoundtype = _get_compound_type_from_kind(kind)
            if compoundtype is None:
                reporter.xml_assert(xmlpath, "unknown compound kind '{0}'".format(kind))
                continue
            compound = compoundtype(name, refid)
            compound.set_documentation_set(self)
            self._compounds[refid] = compound
            for memberelem in compoundelem.iter("member"):
                name = memberelem.find("name").text
                refid = memberelem.attrib["refid"]
                kind = memberelem.attrib["kind"]
                if refid in self._members:
                    member = self._members[refid]
                    membertype = _get_member_type_from_kind(kind)
                    if not isinstance(member, membertype):
                        reporter.xml_assert(
                            xmlpath,
                            "id '{0}' used for multiple kinds of members".format(refid),
                        )
                        continue
                else:
                    membertype = _get_member_type_from_kind(kind)
                    if membertype is None:
                        reporter.xml_assert(
                            xmlpath, "unknown member kind '{0}'".format(kind)
                        )
                        continue
                    member = membertype(name, refid)
                    member.set_documentation_set(self)
                    self._members[refid] = member
                member.add_parent_compound(compound)
                compound.add_member(member)

    def load_file_details(self, filelist=None):
        """Load detailed XML files for all files and possible parents of files.

        If filelist is set, it should be a list of file paths, and details will
        be loaded only for files in those paths.  The paths should be relative
        to the root of the Gromacs source tree."""
        for compound in self._compounds.values():
            if isinstance(compound, (Directory, Group)):
                compound.load_details()
            elif not filelist and isinstance(compound, File):
                compound.load_details()
                self._files[compound.get_path()] = compound
        if filelist:
            # We can't access the full path from the File object before the
            # details are loaded, because Doxygen does not write that into
            # index.xml.  But we can use the Directory objects (where the name
            # is the relative path) to get the path.
            for compound in self._compounds.values():
                if isinstance(compound, File):
                    dirobj = compound.get_directory()
                    if not dirobj:
                        continue
                    relpath = compound.get_directory().get_name()
                    relpath = os.path.join(relpath, compound.get_name())
                    if relpath in filelist:
                        compound.load_details()
                        self._files[compound.get_path()] = compound

    def load_details(self):
        """Load detailed XML files for each compound."""
        for compound in self._compounds.values():
            compound.load_details()
            if isinstance(compound, File):
                self._files[compound.get_path()] = compound
        # TODO: Add links to files using location

    def merge_duplicates(self):
        """Merge duplicate member definitions based on body location.

        At least for some functions that are declared in a header, but have
        their body in a source file, Doxygen seems to create two different IDs,
        but the contents of the members are the same, except for the location
        attribute.  This method merges members that have identical name and
        body location into a single member that keeps the information from both
        instances (they should only differ in the location attribute and in
        parent compounds).  Both IDs point to the merged member after this
        method.
        """
        members_by_body = dict()
        for member in self._members.values():
            bodyloc = member.get_body_location()
            if bodyloc:
                index = (bodyloc, type(member), member.get_name())
                if index not in members_by_body:
                    members_by_body[index] = []
                members_by_body[index].append(member)
        for memberlist in members_by_body.values():
            if len(memberlist) > 1:
                declaration = None
                otherdeclarations = []
                definition = None
                for member in memberlist:
                    if member.has_same_body_location():
                        if definition is not None:
                            self._reporter.xml_assert(
                                None,
                                "duplicate definition for a member '{0}'".format(
                                    definition
                                ),
                            )
                            continue
                        definition = member
                    elif declaration is None:
                        declaration = member
                    else:
                        otherdeclarations.append(member)
                if otherdeclarations:
                    # TODO: gmx_cpuid.c produces some false positives
                    details = []
                    for otherdeclaration in otherdeclarations:
                        details.append(
                            "{0}: another declaration is here".format(
                                otherdeclaration.get_reporter_location()
                            )
                        )
                    details.append(
                        "{0}: definition is here".format(
                            declaration.get_body_location()
                        )
                    )
                    text = "duplicate declarations for a member '{0}'".format(
                        declaration
                    )
                    self._reporter.code_issue(declaration, text, details)
                    continue
                self._members[definition.get_id()] = declaration
                declaration.merge_definition(definition)
                for compound in definition.get_parent_compounds():
                    compound.replace_member(definition, declaration)

    def get_reporter(self):
        """Return reporter object to use for reporting issues.

        This method is used in the entity classes to access the reporter when
        they are parsing the XML files.
        """
        return self._reporter

    def get_xmlroot(self):
        """Return root of the Doxygen XML directory."""
        return self._xmlroot

    def get_compound(self, refid):
        return self._compounds[refid]

    def get_member(self, refid):
        return self._members[refid]

    def get_compounds(self, types, predicate=None):
        result = []
        for compound in self._compounds.values():
            if isinstance(compound, types) and (
                predicate is None or predicate(compound)
            ):
                result.append(compound)
        return result

    def get_members(self, types=None, predicate=None):
        # self._members can contain duplicates because of merge_duplicates()
        result = set()
        for member in self._members.values():
            if (types is None or isinstance(member, types)) and (
                predicate is None or predicate(member)
            ):
                result.add(member)
        return list(result)

    def get_files(self, paths=None):
        if paths:
            return self.get_compounds(File, lambda x: x.get_name().endswith(paths))
        else:
            return self.get_compounds(File)

    def get_directories(self, paths):
        return self.get_compounds(Directory, lambda x: x.get_name().endswith(paths))

    def get_groups(self, name):
        return self.get_compounds(Group, lambda x: x.get_name() in name)

    def get_namespaces(self, name=None):
        if name:
            return self.get_compounds(Namespace, lambda x: x.get_name() in name)
        else:
            return self.get_compounds(Namespace)

    def get_classes(self, name=None):
        if name:
            return self.get_compounds(Class, lambda x: x.get_name() in name)
        else:
            return self.get_compounds(Class)

    def get_functions(self, name):
        return self.get_members(Member, lambda x: x.get_name() in name)


#####################################################################
# Code for running in script mode


def main():
    """Run the script in for debugging/Doxygen XML output inspection."""
    import sys

    from optparse import OptionParser

    from reporter import Reporter

    parser = OptionParser()
    parser.add_option("-R", "--root-dir", help="Doxygen XML root directory")
    parser.add_option(
        "-F", "--show-file", action="append", help="Show contents of given file"
    )
    parser.add_option(
        "-d", "--show-dir", action="append", help="Show contents of given directory"
    )
    parser.add_option(
        "-g", "--show-group", action="append", help="Show contents of given group"
    )
    parser.add_option(
        "-n",
        "--show-namespace",
        action="append",
        help="Show contents of given namespace",
    )
    parser.add_option(
        "-c", "--show-class", action="append", help="Show contents of given class"
    )
    # TODO: Add option for other types, and make them work
    parser.add_option(
        "-f", "--show-function", action="append", help="Show details of given function"
    )
    options, args = parser.parse_args()

    reporter = Reporter()

    sys.stderr.write("Loading index.xml...\n")
    docset = DocumentationSet(options.root_dir, reporter)
    reporter.write_pending()
    sys.stderr.write("Loading details...\n")
    docset.load_details()
    reporter.write_pending()
    sys.stderr.write("Processing...\n")
    docset.merge_duplicates()
    reporter.write_pending()

    objlist = []
    if options.show_file:
        objlist.extend(docset.get_files(tuple(options.show_file)))
    if options.show_dir:
        objlist.extend(docset.get_directories(tuple(options.show_dir)))
    if options.show_group:
        objlist.extend(docset.get_groups(tuple(options.show_group)))
    if options.show_namespace:
        # TODO: Replace file names with anonymous_namespace{filename}
        objlist.extend(docset.get_namespaces(tuple(options.show_namespace)))
    if options.show_class:
        objlist.extend(docset.get_classes(tuple(options.show_class)))
    if options.show_function:
        objlist.extend(docset.get_functions(tuple(options.show_function)))
    for obj in objlist:
        obj.show()


if __name__ == "__main__":
    main()



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2015- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

# -*- coding: utf-8 -*-
#
# GROMACS documentation build configuration file, created by
# sphinx-quickstart on Tue Jan 13 14:28:44 2015.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime
import os
import sys

# The following definitions are completed via CMake machinery.

gmxapi_staging_path = "@GMXAPI_PYTHON_STAGING_DIR@"
if gmxapi_staging_path and os.path.isdir(gmxapi_staging_path):
    sys.path.append(gmxapi_staging_path)

try:
    import gmxapi
except ImportError:
    gmxapi_version_string = "N/A"
else:
    gmxapi_version_string = gmxapi.__version__

gmx_admin_scripts_path = "@GMX_ADMIN_DIR@"
gmx_containers_path = "@GMX_ADMIN_DIR@/containers"
gmx_sphinx_extension_path = "@SPHINX_EXTENSION_PATH@"
releng_path = "@RELENG_PATH@"
gmx_version_string = "@GMX_VERSION_STRING@"
gmx_version_string_full = "@GMX_VERSION_STRING_FULL@"
regressiontest_version = "@REGRESSIONTEST_VERSION@"
gmx_min_sphinx = "@EXPECTED_SPHINX_VERSION@"
variables = [
    ("EXPECTED_DOXYGEN_VERSION", "@EXPECTED_DOXYGEN_VERSION@"),
    ("EXPECTED_SPHINX_VERSION", "@EXPECTED_SPHINX_VERSION@"),
    ("CMAKE_MINIMUM_REQUIRED_VERSION", "@CMAKE_MINIMUM_REQUIRED_VERSION@"),
    ("GMX_CLANG_MINIMUM_REQUIRED_VERSION", "@GMX_CLANG_MINIMUM_REQUIRED_VERSION@"),
    ("GMX_GCC_MINIMUM_REQUIRED_VERSION", "@GMX_GCC_MINIMUM_REQUIRED_VERSION@"),
    ("GMX_CUDA_MINIMUM_REQUIRED_VERSION", "@GMX_CUDA_MINIMUM_REQUIRED_VERSION@"),
    (
        "GMX_CUDA_MINIMUM_REQUIRED_COMPUTE_CAPABILITY",
        "@GMX_CUDA_MINIMUM_REQUIRED_COMPUTE_CAPABILITY@",
    ),
    ("GMX_HDF5_MINIMUM_REQUIRED_VERSION", "@GMX_HDF5_MINIMUM_REQUIRED_VERSION@"),
    ("REQUIRED_OPENCL_MIN_VERSION", "@REQUIRED_OPENCL_MIN_VERSION@"),
    ("SOURCE_MD5SUM", "@SOURCE_MD5SUM@"),
    ("REGRESSIONTEST_MD5SUM", "@REGRESSIONTEST_MD5SUM_STRING@"),
    ("GMX_TNG_MINIMUM_REQUIRED_VERSION", "@GMX_TNG_MINIMUM_REQUIRED_VERSION@"),
    ("GMX_LMFIT_REQUIRED_VERSION", "@GMX_LMFIT_REQUIRED_VERSION@"),
    ("GMX_MANUAL_DOI_STRING", "@GMX_MANUAL_DOI_STRING@"),
    ("GMX_SOURCE_DOI_STRING", "@GMX_SOURCE_DOI_STRING@"),
    ("GMX_CURRENT_CONTRIBUTORS", "@GMX_CURRENT_CONTRIBUTORS_STRING@"),
    ("GMX_PREVIOUS_CONTRIBUTORS", "@GMX_PREVIOUS_CONTRIBUTORS_STRING@"),
    ("GMX_CURRENT_PROJECT_LEADERS", "@GMX_CURRENT_PROJECT_LEADERS_STRING@"),
]
# End of build-time substitutions.

sys.path.extend(
    (gmx_admin_scripts_path, gmx_containers_path, gmx_sphinx_extension_path)
)
if releng_path and os.path.isdir(releng_path):
    sys.path.append(releng_path)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# gmx_min_sphinx is set from the expected minimum version of Sphinx
# in CMakeLists.txt
needs_sphinx = gmx_min_sphinx

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    # The viewcode extension requires extra configuration or consideration to
    # avoid collecting sources for external packages (through intersphninx) or
    # generating broken links for compiled extension modules that do not have
    # inspectable 'code' member data.
    # Ref: https://www.sphinx-doc.org/en/master/usage/extensions/viewcode.html
    # Please run linkchecker and inspect the generated
    # docs/html/_modules/index.html page before committing a change that enables
    # 'sphinx.ext.viewcode',
    "sphinx_copybutton",
    "sphinx_inline_tabs",
    "sphinxarg.ext",
    "sphinxcontrib.autoprogram",
    "gmxsphinx",
]
extlinks = {"issue": ("https://gitlab.com/gromacs/gromacs/-/issues/%s", "Issue %s")}

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# General information about the project.
project = "GROMACS"
copyright = str(datetime.datetime.now().year) + ", GROMACS development team"
thisyear_string = str(datetime.datetime.now().year)
# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = gmx_version_string
# The full version, including alpha/beta/rc tags.
release = gmx_version_string_full
# default file extension for plots
plotext = ".eps"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["fragments"]
# Note: the automatically generated builder tags are not available until after
# conf.py is processed, so we use additional custom tags in our sphinx targets.
if not tags.has("do_man"):
    exclude_patterns += ["man"]
if tags.has("do_latex"):
    exclude_patterns.extend(
        [
            "index.rst",
        ]
    )
    # The master toctree document.
    root_doc = "texindex"
else:
    exclude_patterns += ["texindex.rst"]
    # The master toctree document.
    root_doc = "index"

# Set tag if documentation build can convert images or not
# to selectively include files in reference manual
tags.add("@IMAGE_CONVERT_STRING@")

if tags.has("gmx_image_convert_impossible"):
    exclude_patterns.extend(
        [
            "reference-manual/algorithms",
            "reference-manual/functions",
            "reference-manual/special",
            "reference-manual/analysis",
        ]
    )

# The reST default role (used for this markup: `text`) to use for all
# documents.
default_role = "any"

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
# keep_warnings = False

# Configure the values for all the variables that might later configure any of the .rst files.
substitutions = [".. |{0}| replace:: {1}".format(*x) for x in variables if x[1]]
substitutions.extend(
    [".. |{0}| replace:: unknown".format(x[0]) for x in variables if x[1] == ""]
)
rst_epilog = "\n".join(substitutions)
rst_epilog += """
.. |Gromacs| replace:: GROMACS
.. _gmx-manual: manual-{gmx_version_string}.pdf
.. _gmx-manual-parent-dir: ../manual-{gmx_version_string}.pdf
.. |gmx-source-package-ftp| replace:: As ftp ftp://ftp.gromacs.org/gromacs/gromacs-{gmx_version_string}.tar.gz
.. |gmx-source-package-http| replace:: As https https://ftp.gromacs.org/gromacs/gromacs-{gmx_version_string}.tar.gz
.. |gmx-regressiontests-package| replace:: https://ftp.gromacs.org/regressiontests/regressiontests-{regressiontest_version}.tar.gz
.. |gmxapi-version| replace:: {gmxapi_version_string}
.. _up-to-date installation instructions: https://manual.gromacs.org/documentation/current/install-guide/index.html
.. _developer discussion forum: https://gromacs.bioexcel.eu/c/gromacs-developers/10
.. _user discussion forum: https://gromacs.bioexcel.eu/c/gromacs-user-forum/5
.. _gmx-developers mailing list archive: https://mailman-1.sys.kth.se/pipermail/gromacs.org_gmx-developers
.. _gmx-users mailing list archive: https://mailman-1.sys.kth.se/pipermail/gromacs.org_gmx-users
.. _CUDA: https://developer.nvidia.com/cuda-zone
.. _OpenCL: https://www.khronos.org/opencl/
.. _SYCL: https://www.khronos.org/sycl/
.. _Intel oneAPI DPC++: https://www.intel.com/content/www/us/en/developer/tools/oneapi/dpc-compiler.html
.. _AdaptiveCpp: https://github.com/AdaptiveCpp/AdaptiveCpp
.. _ROCm: https://rocm.docs.amd.com/en/latest/index.html
.. _rocPrim: https://rocm.docs.amd.com/projects/rocPRIM/en/latest/index.html
.. _HDF5: https://www.hdfgroup.org/solutions/hdf5
.. _H5MD: https://h5md.nongnu.org
.. _OpenMPI: http://www.open-mpi.org
.. _MPICH: http://www.mpich.org
.. _OpenMP: http://en.wikipedia.org/wiki/OpenMP
.. _CMake installation page: http://www.cmake.org/install/
.. _Ubuntu toolchain ppa page: https://launchpad.net/~ubuntu-toolchain-r/+archive/ubuntu/test
.. _EPEL page: https://fedoraproject.org/wiki/EPEL
.. _running CMake: http://www.cmake.org/runningcmake/
.. _CMake environment variables: https://gitlab.kitware.com/cmake/community/-/wikis/doc/cmake/Useful-Variables
.. _FFTW: http://www.fftw.org
.. _FFTW installation guide: http://www.fftw.org/doc/Installation-and-Customization.html#Installation-and-Customization
.. _MKL: https://software.intel.com/en-us/intel-mkl
.. _VMD: http://www.ks.uiuc.edu/Research/vmd/
.. _PyMOL: http://www.pymol.org
.. _webpage: http://www.gromacs.org
.. _ftp site: ftp://ftp.gromacs.org/gromacs/
.. _third-party-tutorials: http://www.mdtutorials.com/gmx/
.. _tutorials: https://tutorials.gromacs.org/
.. _issue tracker: https://gitlab.com/gromacs/gromacs/-/issues/
.. _gitlab: https://gitlab.com/gromacs/gromacs/
.. _download: ../download.html
.. |thisyear| replace:: {thisyear_string}
""".format(
    gmx_version_string=gmx_version_string,
    gmxapi_version_string=gmxapi_version_string,
    regressiontest_version=regressiontest_version,
    thisyear_string=thisyear_string,
)

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"
# html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    # Hide project title in sidebar, since we are using a logo instead.
    "sidebar_hide_name": True,
    # Add an "edit" button to the rendered web pages.
    "source_repository": "https://gitlab.com/gromacs/gromacs/",
    "source_branch": "main",
    "source_directory": "docs/",
}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [""]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "GROMACS " + release + " documentation"

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = "GROMACS " + version

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "reference-manual/plots/GMX_logos/gmx_logo_blue.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
# html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "Gromacsdoc"


# -- Options for LaTeX output ---------------------------------------------
# Note: the automatically generated builder tags are not available until after
# conf.py is processed, so we use additional custom tags in our sphinx targets.
if tags.has("do_latex"):
    latex_elements = {
        # The paper size ('letterpaper' or 'a4paper').
        "papersize": "a4paper",
        # The font size ('10pt', '11pt' or '12pt').
        "pointsize": "11",
        # Additional stuff for the LaTeX preamble.
        # The tocdepth setting is needed to overwrite the default value set by Sphinx
        # to get a more detailed toctree in the pdf version of the manual.
        "preamble": r"""
        \usepackage{here}
        \usepackage{picins}
        \usepackage{underscore}
        \usepackage{tabularx}
        \usepackage{multicol}
        \usepackage{dcolumn}
        \usepackage{makeidx}
        \usepackage{times}
        \usepackage{ifthen}
        \usepackage{enumitem}
        \usepackage{longtable}
        \usepackage{pdflscape}
        \pagenumbering{roman}
        \usepackage{array}
        \setcounter{tocdepth}{2}
        """,
        # Format output to avoid empty pages
        "classoptions": ",openany,oneside",
    }

    # Grouping the document tree into LaTeX files. List of tuples
    # (source start file, target name, title,
    #  author, documentclass [howto, manual, or own class]).
    latex_documents = [
        (
            "texindex",
            "gromacs.tex",
            "GROMACS Documentation",
            "GROMACS development team",
            "manual",
        ),
    ]

    # The name of an image file (relative to this directory) to place at the top of
    # the title page.
    latex_logo = "reference-manual/plots/GMX_logos/gmx_falcon_blue.png"

    # For "manual" documents, if this is true, then toplevel headings are parts,
    # not chapters.
    latex_use_parts = True

    # If true, show page references after internal links.
    latex_show_pagerefs = True

    # If true, show URL addresses after external links.
    # latex_show_urls = False

    # Documents to append as an appendix to all manuals.
    # latex_appendices = []

    # If false, no module index is generated.
    # latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# Note: the automatically generated builder tags are not available until after
# conf.py is processed, so we use additional custom tags in our sphinx targets.
if tags.has("do_man"):
    # Note: conf-man.py is generated by the (built) command line executable.
    # Ref: gmx::CommandLineHelpModule
    exec(open("conf-man.py").read())

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output -------------------------------------------
# Note: the automatically generated builder tags are not available until after
# conf.py is processed, so we use additional custom tags in our sphinx targets.
if tags.has("do_texinfo"):

    # Grouping the document tree into Texinfo files. List of tuples
    # (source start file, target name, title, author,
    #  dir menu entry, description, category)
    texinfo_documents = [
        (
            "texindex",
            "GROMACS",
            "GROMACS Documentation",
            "GROMACS development team",
            "GROMACS",
            "One line description of project.",
            "Miscellaneous",
        ),
    ]

    # Documents to append as an appendix to all manuals.
    # texinfo_appendices = []

    # If false, no module index is generated.
    # texinfo_domain_indices = True

    # How to display URL addresses: 'footnote', 'no', or 'inline'.
    # texinfo_show_urls = 'footnote'

    # If true, do not generate a @detailmenu in the "Top" node's menu.
    # texinfo_no_detailmenu = False

# Make it possible to use numbered labels for figures and tables
numfig = True

# -- Options for autodoc extension ----------------------------------------

# http://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autodoc-mock-imports
autodoc_mock_imports = ["hpccm", "hpccm.config"]

# -- Options for intersphinx extension ------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "mpi4py": ("https://mpi4py.readthedocs.io/en/stable/", None),
}
intersphinx_cache_limit = -1
intersphinx_timeout = 10



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Manage computing resources and runtime context.

Utilities, context managers, and singletons for handling resource allocation
and lifetime management.

For the purposes of this module, the term "assignment" refers to resources that
have been reserved (usually for exclusive use) within a well-defined scope
(such as a specific function call or phase of program execution).

"Allocation" can be taken to mean "resources that are available to be assigned,"
and may or may not be tightly scoped.

Resource assignment may be nested, in which case an Assignment with a broader
scope in the program is used as the Allocation from which a more local
Assignment is made.

Assignment and Allocation are *roles* that can be served by Context objects.
Abstractly, a Context is a notion of some aspect of program or resource *state*.
A Context may be represented by a concrete class when an object manages the
details needed to participate in a stateful protocol.

Note that resource assignment is usually a collective operation within the
scope of an allocation. For instance, an ``MPI_Comm_split`` call must be made
(synchronously) on all members of the parent MPI communicator. Similar care must
be taken with the Allocation and Assignment protocols in this module.

.. versionadded:: 0.4.0

    This module generalizes some of the resource management previously in
    :py:mod:`gmxapi.simulation.mdrun`, decoupling MPI communicator handling
    from :py:mod:`gmxapi.simulation.context`.
    See also https://gitlab.com/gromacs/gromacs/-/issues/3718

"""

__all__ = (
    "filtered_mpi_environ",
    "filtered_prefixes",
    "scoped_resources",
    "ResourceAllocation",
    "ResourceAssignment",
    "ResourceRequirements",
)

import dataclasses
import functools
import os
import threading
import typing
import warnings
import weakref
from contextlib import contextmanager

import gmxapi.utility
from gmxapi import exceptions
from gmxapi import logger as root_logger

try:
    from typing import Protocol
except ImportError:
    # typing.Protocol provides helpful static type checking support for
    # interfaces, but is not availbale until Python 3.8.
    # noinspection PyTypeChecker
    Protocol = typing.Generic

# Initialize module-level logger
logger = root_logger.getChild(__name__)
logger.info("Importing {}".format(__name__))


if typing.TYPE_CHECKING:
    try:
        import mpi4py.MPI.Comm as Communicator
    except ImportError:
        Communicator = None


# gmxapi requires mpi4py, but it is useful to allow module-level imports to fail
# so that documentation builds can succeed without package dependencies. It is
# also helpful to retain the accompanying logic in case we want to decouple from
# mpi4py at some point in the future.
@functools.lru_cache()
def mpi_comm_world():
    from mpi4py.MPI import COMM_WORLD

    return COMM_WORLD


@functools.lru_cache()
def mpi_comm_null():
    from mpi4py.MPI import COMM_NULL

    return COMM_NULL


@functools.lru_cache()
def mpi_undefined():
    from mpi4py.MPI import UNDEFINED

    return UNDEFINED


_context_lock = threading.Lock()
"""Use to avoid race conditions when changing singleton state."""

CommunicatorT = typing.TypeVar("CommunicatorT")
ProvidesTokenT = typing.TypeVar("ProvidesTokenT")
ReceivedTokenT = typing.TypeVar("ReceivedTokenT")


class ResourceAllocation(Protocol[CommunicatorT, ProvidesTokenT]):
    """The interface we expect for a ResourceAllocation role."""

    def communicator(self) -> typing.Optional[CommunicatorT]:
        """Get a reference to the communicator allocated for the current task scope, if any."""
        raise NotImplementedError

    def subscribe(self, subscriber) -> ProvidesTokenT:
        """Alert the ResourceAllocation that its resources are in use.

        For initial implementation, we cannot support locked resources with
        an ``acquire()`` / ``release()`` protocol. But we can still support
        some bookkeeping and issue warnings for developers.
        """
        raise NotImplementedError

    def unsubscribe(self, token: ProvidesTokenT):
        """Notify the ResourceAllocation that a subscriber is done using resources."""
        raise NotImplementedError


def _finalize_communicator(comm):
    if comm == mpi_comm_null():
        return
    # noinspection PyBroadException
    try:
        comm.Free()
    except Exception:
        logger.exception(f"Could not Free {comm}.")


class ResourceAssignment(
    ResourceAllocation[CommunicatorT, int],
    typing.Generic[CommunicatorT, ReceivedTokenT],
):
    """Container for assigned resources.

    If an MPI communicator is provided when initializing the ResourceAssignment,
    ownership of the communicator is **not** assumed, and `comm.Free()` will
    **not** be called automatically when the ResourceAssignment is closed or
    finalized.

    To transfer ownership of the communicator in *task* or *ensemble*, add
    a finalizer using either the *finalizer* key word argument or the
    `add_finalizer` method.
    """

    parent: "weakref.ReferenceType[ResourceAllocation[CommunicatorT, ReceivedTokenT]]"
    """Resource allocation from which the current assignment is derived.
    
    Notes that this is a weak reference to the provided ResourceAllocation.
    See :py:mod:`weakref`.
    """

    _resource_id: typing.Optional = None
    """Global identifier for this assignment.
    
    If multiple threads or processes share the assignment, the identifier
    is shared across the threads and processes.

    Across threads or processes, multiple instances of ResourceAssignment
    could refer to the same (multi-core) assignment. This identifier can be
    used to identify callers sharing a resource assignment for a particular
    scope of execution. Callers may be grouped by task (within an ensemble)
    using *_group_id*, and within a task using *_local_id* (such as to
    identify the "root" participant in a task).
    
    This attribute is experimental. If used, it is the callers responsibility
    to set the value uniformly.
    """

    _group_ids: typing.Optional[typing.Tuple[int]] = None
    """The resource subgroup within *_resource_id*.

    This identifier is not generally unique when *_resource_id* varies.
    The identifier _may_ be set even for non-participating callers to indicate
    that resources owned by the caller are being used for the task associated
    with the group.
    
    This attribute is experimental. If used, it is the callers responsibility
    to set the value appropriately.    
    """

    _roots: typing.Optional[typing.Tuple[typing.Union[int, None]]] = None
    """Ensemble rank identifiers for subtask roots.
    
    Generally equal to the _group_ids for participants that are subtask roots
    and None otherwise.
    
    This information could be stored in a more concise structure, but a little
    redundancy is useful for error checking and it is convenient to have the
    same look-up scheme of ``attr[_base_comm_rank]``.
    
    This attribute is experimental. If used, it is the callers responsibility
    to set the value appropriately.    
    """

    _subtask_ranks: typing.Optional[typing.Tuple[typing.Union[int, None]]] = None
    """Instance identifier for this assignment.
    
    Generally, serves as the "participant ID" for a resource assignment
    comprising multiple threads or processes.
    
    This identifier is not generally unique when *_resource_id* varies or
    when *_group_id* varies. The identifier should be `None` for
    non-participating callers.

    This attribute is experimental. If used, it is the callers responsibility
    to set the value with appropriate uniqueness.
    """

    _token: ReceivedTokenT
    _subscribers: list

    __base_comm_rank: typing.Optional[int] = None
    """Rank of the caller in the parent allocation."""

    __communicator: typing.Optional[CommunicatorT]
    __ensemble: typing.Optional[CommunicatorT]
    __finalizers: list
    __next_monotonic_integer = 0

    def __init__(
        self,
        *,
        subtask: CommunicatorT,
        parent: ResourceAllocation[CommunicatorT, ReceivedTokenT],
        ensemble: CommunicatorT = None,
        finalizers: typing.Sequence[weakref.finalize] = (),
    ):
        try:
            self.__base_comm_rank = parent.communicator().Get_rank()
            self._base_comm_size = parent.communicator().Get_size()
            self.parent = weakref.ref(parent)
        except (AttributeError, TypeError) as e:
            raise ValueError(f"Unusable resources: {parent}.") from e
        self.__communicator = subtask
        self.__ensemble = ensemble
        try:
            self._token = parent.subscribe(self)
            logger.debug(f"{self} subscribed to {parent}. Token: {self._token}")
        except AttributeError:
            self._token = None
        self._subscribers = list()
        self.__finalize_dirty = weakref.finalize(
            self,
            warnings.warn,
            message=f"{repr(self)} finalized without an explicit call to close().",
            category=ResourceWarning,
        )
        self.__finalizers = [self.__finalize_dirty]
        if self._token is not None:
            self.__finalizers.append(
                weakref.finalize(self, parent.unsubscribe, self._token)
            )
        self.__finalizers.append(
            weakref.finalize(
                self, _remaining_subscribers_warning, repr(self), self._subscribers
            )
        )
        self.__finalizers.extend(finalizers)

    def __repr__(self):
        return (
            f"<ResourceRequirements:{self.__base_comm_rank} _resource_id:{self._resource_id} _group_ids:"
            f"{self._group_ids} _roots:{self._roots} _subtask_ranks:{self._subtask_ranks} parent:"
            f"{self.parent}:{self._token} communicator:{self.__communicator} ensemble:{self.__ensemble}>"
        )

    def _next_monotonic_integer(self):
        i = self.__next_monotonic_integer
        self.__next_monotonic_integer = i + 1
        return i

    def subscribe(self, subscriber) -> int:
        """Receive a subscription notification.

        Implements ResourceAllocation.
        """
        _id = id(subscriber)
        _repr = repr(subscriber)
        token = self._next_monotonic_integer()
        self._subscribers[token] = (_id, _repr)
        return token

    def unsubscribe(self, token: int):
        """Receive an unsubscription notification.

        Implements ResourceAllocation.
        """
        del self._subscribers[token]

    def communicator(self) -> CommunicatorT:
        """Get the communicator for the current process.

        In an MPI context, other Python interpreter processes participating in
        the current work may have different communicators or null communicators.
        """
        return self.__communicator

    def parent_communicator(self) -> typing.Optional[CommunicatorT]:
        """Get a reference to the communicator from which these resources are derived."""
        parent = self.parent()
        if parent:
            return parent.communicator()
        else:
            return None

    def ensemble_communicator(self) -> typing.Optional[CommunicatorT]:
        """Get a reference to the communicator that connects the members of an ensemble task."""
        return self.__ensemble

    def close(self):
        if self.closed:
            logger.debug(f"{self}.close() called on an already-closed resource.")
        else:
            self.__finalize_dirty.detach()
            for finalizer in reversed(self.__finalizers):
                finalizer()
            del self.__finalizers
            logger.debug(f"Closed {self}.")

    def add_finalizer(self, func, *args, **kwargs):
        if self.closed:
            raise exceptions.UsageError(
                "Cannot add finalizer to a resource that is already closed."
            )
        self.__finalizers.append(weakref.finalize(self, func, *args, **kwargs))

    @property
    def closed(self):
        return not self.__finalize_dirty.alive

    @classmethod
    def assign(
        cls,
        allocation: ResourceAllocation[CommunicatorT, ReceivedTokenT],
        requirements: "ResourceRequirements",
    ):
        """Assign resources to members of a new ensemble task.

        If *allocation* includes an MPI communicator, the resources yielded will
        contain a sub-communicator of *allocation.communicator()*.

        Args:
            allocation: available resources, or *None* to determine automatically.
            requirements: parameters for the resource assignment, such as MPI communicators.

        *allocation* is required if *requirements* is not *None*. *requirements*
        may be validated in terms of the provided *allocation*.

        When GROMACS is not built with an MPI library, the number of ranks used
        is equal to *len(requirements.communication)*. Participating ranks are
        not necessarily consecutive.

        Specifically, ``ResourceAllocation.assign(...).communicator`` will be
        :py:class:`~mpi4py.MPI.COMM_NULL` in some processes
        (consistent with :py:func:`mpi4py.MPI.Comm.Split()`) when
        - ``mpi4py`` is in use (``gmxapi.version.has_feature('mpi_bindings') == True``
          or *allocation.communicator()* is an :py:class:`mpi4py.MPI.Comm`),
        - ``gmx.utility.config()['gmx_mpi_type'] != 'library'``,
        - *requirements* is not ``None``, and
        - ``len(requirements.communication) < allocation.communicator().Get_size()``.

        Callers should inspect the yielded resources to decide whether to
        use the new communicator.

        For MPI-enabled GROMACS simulators
        (``gmx.utility.config()['gmx_mpi_type'] == 'library'``), all available
        ranks in *allocation.communicator()* are distributed as evenly as possible
        to the new subcommunicator(s) in contiguous blocks.
        """
        if requirements is not None:
            if allocation is None:
                raise exceptions.UsageError(
                    "Provide allocated resources from which to request a sub-allocation."
                )

        if allocation is None:
            # We need to think more about race conditions and other structural
            # details before allowing implicit automatic detection of resources.
            raise exceptions.FeatureNotAvailableError(
                "Implicit allocation is not currently available."
            )
        base_communicator = allocation.communicator()
        base_comm_size = base_communicator.Get_size()
        base_comm_rank = base_communicator.Get_rank()

        n_subtasks = len(requirements.communication)
        subtask_comm = None
        ensemble_comm = None
        _resource_id = None
        group_ids = None
        _local_id = None
        local_ids = None
        roots = None
        if n_subtasks > 0:
            if n_subtasks > base_comm_size:
                error = f"Cannot satisfy {requirements} from {allocation}."
                logger.error(
                    error
                    + " "
                    + "To get a scoped communicator of a specific size, provide a sufficient "
                    + "communicator supporting the mpi4py.MPI.Comm interface."
                )
                raise exceptions.UsageError(error)

            group_ids = tuple(
                _rank * n_subtasks // base_comm_size for _rank in range(base_comm_size)
            )
            """Label a number of contiguous blocks of processes in an allocation.
    
            Divide the ranks of *allocation* into *count* consecutive blocks of
            (roughly) equal size, labeled from 0 to (*count* - 1). Get a list
            corresponding to the label for each rank.
        
            Warning:
                Blocks may span node boundaries.
                When allocation spans multiple nodes, but blocks are smaller than one node,
                we should try to align blocks at node boundaries for more uniform
                communication latency. This is not currently possible (but behavior
                should be kept consistent with `roots`).
            """

            assert set(group_ids) == set(range(n_subtasks))
            _group_id = group_ids[base_comm_rank]

            # Get the index of the first occurrence of each group ID.
            root_indices = set(group_ids.index(i) for i in range(n_subtasks))
            roots = tuple(
                group_ids[_rank] if _rank in root_indices else None
                for _rank in range(base_comm_size)
            )
            """Sparse assignment for each rank in allocation.
            
            Integers from 0 to (*count* - 1) will be distributed in a sequence with
            as many elements as the communicator size in *allocation*. Only one rank
            from the communicator is associated with each ID. Other ranks are
            labeled `None`.
        
            Warning:
                Blocks may span node boundaries.
                When allocation spans multiple nodes, we should try to align blocks at
                node boundaries so that we can assume that a rank assigned None could
                lend its CPU core to the closest non-None member. This is not currently
                possible. If pursued in the future, it could be appropriate to create
                an additional resource layer comprising a parent MPI.Group for the
                processes "owning" each subset of cores.

            """

            assert len(set(roots) - {None}) == n_subtasks
            assert set(roots) - {None} == set(range(n_subtasks))

            _current_offset = 0
            subtask_ranks = [0] * base_comm_size
            for _rank in range(base_comm_size):
                if roots[_rank] is not None:
                    _current_offset = _rank
                subtask_ranks[_rank] = _rank - _current_offset

            communication_requirements = tuple(
                key for task in requirements.communication for key in task
            )
            for requirement in communication_requirements:
                if requirement == "subtask_comm":
                    # We don't support any options, but the participants must agree.
                    assert all(
                        subtask.get("subtask_comm", False)
                        for subtask in requirements.communication
                    )

                    if gmxapi.utility.config()["gmx_mpi_type"] == "library":
                        color = tuple(subtask_id for subtask_id in group_ids)
                    else:
                        color = tuple(
                            mpi_undefined() if roots[i] is None else subtask_id
                            for i, subtask_id in enumerate(group_ids)
                        )
                    subtask_comm = base_communicator.Split(
                        color[base_comm_rank], key=subtask_ranks[base_comm_rank]
                    )
                    if subtask_comm:
                        _local_id = subtask_comm.Get_rank()
                        # Check expected behavior for current resource allocation scheme.
                        expected_comm_size = color.count(color[base_comm_rank])
                        actual_comm_size = subtask_comm.Get_size()
                        if gmxapi.utility.config()["gmx_mpi_type"] != "library":
                            assert expected_comm_size == 1
                        assert actual_comm_size == expected_comm_size
                        assert _local_id == subtask_ranks[base_comm_rank]
                elif requirement == "ensemble_comm":
                    # We are currently assuming that all task requirements include
                    # a specifier for the same ensemble.
                    if not all(
                        "ensemble_comm" in task for task in requirements.communication
                    ):
                        raise exceptions.ProtocolError(
                            "All task requirements are assumed to specify the same ensemble."
                        )
                    # We allow the caller to provide arbitrary ensemble labels, but we
                    # do not yet track them for uniqueness. For now, just try to use the
                    # label to produce a reasonably unique MPI_INT.
                    ensemble_labels = set(
                        task["ensemble_comm"] for task in requirements.communication
                    )
                    assert len(ensemble_labels) == 1
                    _resource_id = ensemble_labels.pop()
                    logger.debug(f"Resource ID: {_resource_id}")
                    color = tuple(
                        mpi_undefined() if _id is None else _resource_id
                        for _id in roots
                    )

                    ensemble_comm = base_communicator.Split(color[base_comm_rank])
                else:
                    raise ValueError(
                        f"Unrecognized communication requirement: {requirement}"
                    )

            # One collective MPI call serves as a useful synchronization and error
            # checking point, as well as a convenient way to get the following data.
            local_ids = base_communicator.allgather(_local_id)
            assert len(local_ids) == base_comm_size
            assert local_ids[base_comm_rank] == _local_id

        resources = cls(subtask=subtask_comm, parent=allocation, ensemble=ensemble_comm)
        assert base_comm_rank == resources.base_rank()
        # Note: mpi4py.MPI.Comm.__bool__() compares to MPI_COMM_NULL.
        if subtask_comm:
            resources.add_finalizer(
                weakref.finalize(resources, _finalize_communicator, subtask_comm)
            )
        if ensemble_comm:
            resources.add_finalizer(
                weakref.finalize(resources, _finalize_communicator, ensemble_comm)
            )
        if local_ids is None:
            resources._subtask_ranks = tuple([None] * base_comm_size)
        else:
            resources._subtask_ranks = tuple(local_ids)
        resources._group_ids = group_ids
        resources._resource_id = _resource_id
        resources._roots = roots

        return resources

    def subtask_rank(self):
        if self.__communicator:
            return self._subtask_ranks[self.__base_comm_rank]

    def subtask_id(self):
        """Identifier for the (group of) resource participants.

        Within the scope of a `resource_id`, callers participating
        in the same subtask can be identified by sharing the same subtask ID.

        Returns None for callers that are not participating.
        """
        if self.__communicator:
            return self._group_ids[self.__base_comm_rank]

    def is_subtask_root(self):
        """Whether the participant is the "root" member of the subtask.

        If True, then `subtask_id` is also the rank in the `ensemble_communicator`, if any.
        """
        if self._roots is not None:
            return self._roots[self.__base_comm_rank] is not None

    def base_rank(self):
        """Identify the caller within the parent allocation."""
        return self.__base_comm_rank


def _remaining_subscribers_warning(provider: str, subscribers: list):
    for token, subscriber in subscribers:
        _id, _repr = subscriber
        warnings.warn(
            f"Closing {provider} while {_repr} (id: {_id}, token: {token}) is still subscribed."
        )


class BaseContext:
    """Resource allocation and options.

    Includes default run time execution parameters that may be configured only
    until resources are assigned. That is, class data is only mutable while
    no instance exists.
    """

    __instance: typing.ClassVar[typing.Optional[weakref.ReferenceType]] = None

    __mpi_communicator: typing.ClassVar[typing.Optional["Communicator"]] = None

    def __new__(cls, *args, **kwargs):
        # Check the weakref for a singleton instance, if any.
        if cls.__instance is None:
            instance = None
        else:
            instance = cls.__instance()
        # Set a new singleton instance, if necessary.
        if instance is None:
            instance = super(BaseContext, cls).__new__(cls, *args, **kwargs)
            cls.__instance = weakref.ref(instance)
        return instance

    @classmethod
    def communicator(cls) -> "Communicator":
        return cls.__mpi_communicator

    @classmethod
    def set_base_communicator(cls, communicator):
        if cls.__instance is not None:
            raise exceptions.UsageError(
                "Cannot edit resource defaults after initialization."
            )
        cls.__mpi_communicator = communicator

    @classmethod
    def instance(cls):
        """Get the singleton instance. (Finalize and instantiate if needed.)"""
        if cls.__mpi_communicator is None and mpi_comm_world() is not None:
            cls.set_base_communicator(mpi_comm_world())
        return cls()

    def subscribe(self, subscriber):
        """Accept a subscription notification.

        Supports ResourceAllocation protocol, but has no effect.
        """
        return None

    def unsubscribe(self, token):
        """Accept an un-subscription notification.

        Supports ResourceAllocation protocol, but has no effect.
        """
        return None


# TypedDict is not available until Python 3.8, or by adding `typing-extensions` as a project dependency.
# class _CommunicationRequirements(typing.TypedDict, total=False):
#     ensemble_comm: int
#     subtask_comm: bool
_CommunicationRequirements = dict


@dataclasses.dataclass
class ResourceRequirements:
    """Named requirements for resources to be assigned for a task."""

    communication: typing.Tuple[_CommunicationRequirements, ...] = dataclasses.field(
        default_factory=tuple
    )
    """Requested communicator details.
    
    This field is a container with one element for each block of resources.
    Further details of the container or element type are not yet specified.
    """

    def __post_init__(self):
        for requirement in self.communication:
            if requirement.get("ensemble_comm", 0) // int(2**16) > 0:
                raise ValueError(f"Ensemble communicator labels must fit into uint_16.")


@contextmanager
def scoped_resources(
    allocation: ResourceAllocation["Communicator", typing.Any],
    requirements: ResourceRequirements,
):
    """Manage computing resources with a lifecycle.

    Get a Python context manager for a new resource assignment that is released
    on exiting the `with` block. (Wraps `assign_ensemble()`).

    Args:
        allocation: available resources, or *None* to determine automatically.
        requirements: parameters for the resource assignment, such as MPI communicators.

    *allocation* is required if *requirements* is not *None*. *requirements*
    may be validated in terms of the provided *allocation*.

    .. versionchanged:: 0.4

        Previous versions transformed :py:class:`~mpi4py.MPI.COMM_NULL` into a mock
        communicator before returning for consistency in trivial cases with or
        without :py:mod:`mpi4py`. This translation adds unnecessary indirection,
        and has been removed.

        Similarly, for *original_comm.Get_size()* or *requested_size* of 1,
        an unspecified "fake" communicator could be produced to avoid an
        ``MPI_Comm_split()``. While more efficient, the code complexity was
        problematic and has been removed.

    TODO(#4079): Acquire Task computing resources in a more limited scope than data resources.

        This context manager is called "scoped_resources" instead of "scoped_assignment"
        to avoid implying there is any locking on the usage of the returned resources.
        gmxapi.operation.ResourceManager has a hard-coded protocol for preparing task
        resources that first recursively resolves inputs, then sets up the task runner.
        gmxapi.simulation.mdrun handles communication resources in its input adapter,
        LegacyImplementationSubscription, which ends up being reentrant for chained
        simulations, calling ResourceManager.local_input() _after_ sub-allocating
        Resources. This is a bad design, and it means that there is the appearance
        of resource oversubscription because multiple handles exist to sub-allocations
        from the same base resources. However, there are legitimate reasons for
        components to use multiple communicators on the same ranks for different
        groups of communicating participants, and it is not clear how best to distinguish
        between this sort of usage, either.

    """
    resources = ResourceAssignment.assign(
        allocation=allocation, requirements=requirements
    )

    try:
        yield resources
    finally:
        resources.close()


filtered_prefixes = (
    "DCMF_",  # IBM
    "MPICH_",
    "MPIEXEC_",
    "MPIO_",  # IBM
    "MV2_",  # MVAPICH2 and some forks
    "MVAPICH_",
    "HYDRA_",  # MPICH
    "OMPI_",  # OpenMPI
    "PMI_",  # Process Management Interface
    "PMIX_",  # Newer PMI and batch systems
    "I_MPI_",  # Intel MPI
)
"""MPI-related environment variable prefixes.

Environment variable prefixes known to be associated with MPI implementations,
which may affect MPI context detection, and which should not matter outside of
MPI contexts.

References:
    * :issue:`4423`
    * :issue:`4736`

"""


def filtered_mpi_environ() -> dict:
    """Return a filtered environment variables map with MPI-related entries removed.

    See Also:
        :py:data:`filtered_prefixes`

    """
    return {
        key: value
        for key, value in os.environ.items()
        if not any(key.startswith(prefix) for prefix in filtered_prefixes)
    }



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
"""
The `kinetic_energy` module is part of the physical_validation package, and
consists of checks of the kinetic energy distribution and its
equipartition.
"""
from __future__ import print_function
from __future__ import division

from .util import kinetic_energy as util_kin
from .data import SimulationData


def mb_ensemble(data, alpha=None, verbosity=1,
                screen=False, filename=None):
    r"""Checks if a kinetic energy trajectory is Maxwell-Boltzmann distributed.

    Parameters
    ----------
    data : SimulationData
        Simulation data object
    alpha : float, optional
        If a confidence interval is given and verbose=True, the test outputs
        a passed / failed message.
    verbosity : int, optional
        Verbosity level, where 0 is quiet and 3 very chatty. Default: 1.
    screen : bool, optional
        Plot distributions on screen. Default: False.
    filename : string, optional
        Plot distributions to `filename`.pdf. Default: None.

    Returns
    -------
    result : float
        The p value of the test.

    Notes
    -----
    This function checks whether the hypothesis that a sample
    of kinetic energies is Maxwell-Boltzmann distributed given a specific
    target temperature and the number of degrees of freedom in the system,

    .. math::
        P(K) \sim K^{N/2} e^{-\beta K} \, ,

    holds under a given confidence level :math:`\alpha`.
    The check is performed using the Kolmogorov-Smirnov test provided by
    scipy.stats.kstest_.

    .. _scipy.stats.kstest: https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.stats.kstest.html

    .. note:: The Kolmogorov-Smirnov test is known to have two weaknesses.

       #. The test is more sensitive towards deviations around the center
          of the distribution than at its tails. We deem this to be acceptable
          for most MD applications, but be wary if yours is sensible to the
          kinetic distribution tails.
       #. The test is not valid if its parameters are guessed from the data
          set. Using the target temperature of the MD simulation as an input
          is therefore perfectly valid, but using the average temperature
          over the trajectory as an input to the test can potentially
          invalidate it.

    .. todo:: Can we check the influence of sample size on test results?

    """
    ndof = (data.system.natoms * 3 -
            data.system.nconstraints -
            data.system.ndof_reduction_tra -
            data.system.ndof_reduction_rot)

    return util_kin.check_mb_ensemble(kin=data.observables.kinetic_energy,
                                      temp=data.ensemble.temperature,
                                      ndof=ndof, alpha=alpha,
                                      kb=data.units.kb, verbosity=verbosity,
                                      screen=screen, filename=filename,
                                      ene_unit=data.units.energy_str)


def equipartition(data, dtemp=0.1, distribution=False, alpha=0.05,
                  molec_groups=None,
                  random_divisions=0, random_groups=0,
                  verbosity=2,
                  screen=False, filename=None):
    r"""Checks the equipartition of a simulation trajectory.

    Parameters
    ----------
    data : SimulationData
        Simulation data object
    dtemp : float, optional
        Fraction of temperature deviation tolerated between groups. Default: 0.1 (10%).
    distribution : bool, optional
        If not set, the kinetic energies will not be tested for Maxwell-Boltzmann
        distribution, but only compared amongst each others.
        Default: False.
    alpha : float, optional
        Confidence for Maxwell-Boltzmann test. Default: 0.05 (5%).
    molec_groups : list of array-like (ngroups x ?), optional
        List of 1d arrays containing molecule indeces defining groups. Useful to pre-define
        groups of molecules (e.g. solute / solvent, liquid mixture species, ...). If None,
        no pre-defined molecule groups will be tested. Default: None.

        *Note:* If an empty 1d array is found as last element in the list, the remaining
        molecules are collected in this array. This allows, for example, to only
        specify the solute, and indicate the solvent by giving an empty array.
    random_divisions : int, optional
        Number of random division tests attempted. Default: 0 (random division tests off).
    random_groups : int, optional
        Number of groups the system is randomly divided in. Default: 2.
    verbosity : int, optional
        Verbosity level, where 0 is quiet and 3 very chatty. Default: 2.
    screen : bool
        Plot distributions on screen. Default: False.
    filename : string
        Plot distributions to `filename`.pdf. Default: None.

    Returns
    -------
    result : list
        List of deviations or p-values (if distribution). Tune up verbosity for details.

    Notes
    -----
    This function compares the kinetic energy between groups of degrees of
    freedom. Theoretically, the kinetic energy is expected (via the
    equipartition theorem) to be equally distributed over all degrees of
    freedom. In practice, deviations of temperature between groups of
    degrees of freedom up to several degrees K are routinely observed.
    Larger deviations can, however, hint to misbehaving simulations, such
    as, e.g., frozen degrees of freedom, lack of energy exchange between
    degrees of freedom, and transfer of heat from faster to slower
    oscillating degrees of freedom.

    Splitting of degrees of freedom is done both on a sub-molecular and on
    a molecular level. On a sub-molecular level, the degrees of freedom of
    a molecule can be partitioned into rigid-body contributions
    (translation of the center-of-mass, rotation around the
    center-of-mass) and intra-molecular contributions. On a molecular
    level, the single molecules of the system can be divided in groups,
    either by function (solute / solvent, different species of liquid
    mixtures, ...) or randomly.

    `check_equipartition()` compares the partitioned temperatures of the
    entire system and, optionally, of predefined or randomly separated
    groups.

    Note: In theory, the kinetic energy of the subgroups are expected to
    be individually Maxwell-Boltzmann distributed. As this is seldomly
    holding in practice (see above), `check_equipartition()` is by
    default checking only for abnormal deviations in average temperatures.
    The more strict Maxwell-Boltzmann testing can be invoked by giving the
    flag `distribution`.

    """
    if distribution:
        temp = data.ensemble.temperature
    else:
        temp = None

    (result,
     data.system.ndof_per_molecule,
     data.observables.kinetic_energy_per_molecule) = util_kin.check_equipartition(
        positions=data.trajectory['position'],
        velocities=data.trajectory['velocity'],
        masses=data.system.mass,
        molec_idx=data.system.molecule_idx,
        molec_nbonds=data.system.nconstraints_per_molecule,
        natoms=data.system.natoms,
        nmolecs=len(data.system.molecule_idx),
        ndof_reduction_tra=data.system.ndof_reduction_tra,
        ndof_reduction_rot=data.system.ndof_reduction_rot,
        dtemp=dtemp, temp=temp, alpha=alpha,
        molec_groups=molec_groups,
        random_divisions=random_divisions,
        random_groups=random_groups,
        ndof_molec=data.system.ndof_per_molecule,
        kin_molec=data.observables.kinetic_energy_per_molecule,
        verbosity=verbosity,
        screen=screen,
        filename=filename
    )

    return result



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.
import os

import pytest

import gmxapi as gmx

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    rank_number = comm.Get_rank()
    comm_size = comm.Get_size()
except ImportError:
    comm = None
    rank_number = 0
    comm_size = 1
    rank_tag = ""
    MPI = None
else:
    rank_tag = "rank{}:".format(rank_number)

mpi_support = pytest.mark.skipif(
    comm_size > 1
    and gmx.utility.config()["gmx_mpi_type"] == "library"
    and not gmx.version.has_feature("mpi_comm_integration"),
    reason="Multi-rank MPI contexts require gmxapi 0.4.",
)


@gmx.function_wrapper(output={"data": float})
def add_float(a: float, b: float) -> float:
    return a + b


@gmx.function_wrapper(output={"data": bool})
def less_than(lhs: float, rhs: float, output=None):
    output.data = lhs < rhs


def test_subgraph_function():
    subgraph = gmx.subgraph(variables={"float_with_default": 1.0, "bool_data": True})
    with subgraph:
        # Define the update for float_with_default to come from an add_float operation.
        subgraph.float_with_default = add_float(
            subgraph.float_with_default, 1.0
        ).output.data
        subgraph.bool_data = less_than(
            lhs=subgraph.float_with_default, rhs=6.0
        ).output.data
    operation_instance = subgraph()
    operation_instance.run()
    assert operation_instance.values["float_with_default"] == 2.0

    loop = gmx.while_loop(operation=subgraph, condition=subgraph.bool_data)
    handle = loop()
    assert handle.output.float_with_default.result() == 6


_previous_cpt = [""] * comm_size


@gmx.function_wrapper(output={"data": bool})
def _is_new(current_cpt: str, output):
    # Is the simulation new or is it continuing from a checkpoint?
    global _previous_cpt
    # TODO: Don't use private details of the publishing data proxy.
    # Either use another subgraph variable for previous_cpt, distribute function_wrapper tasks by MPI rank,
    # or develop the interface for session resources provided to tasks.
    executing_member = output._client_identifier
    previous_cpt = _previous_cpt[executing_member]
    _previous_cpt[executing_member] = current_cpt
    if current_cpt != previous_cpt and os.path.exists(previous_cpt):
        output.data = False
    else:
        output.data = True


@mpi_support
def test_subgraph_simulation_extension(spc_water_box, mdrun_kwargs):
    tpr_list = gmx.read_tpr([spc_water_box] * comm_size)
    input_list = gmx.modify_input(tpr_list, parameters={"nsteps": 10**6})
    subgraph = gmx.subgraph(
        variables={
            "new": True,
            "checkpoint": "",
            # 'previous': '',
        }
    )
    with subgraph:
        md = gmx.mdrun(
            input_list,
            runtime_args={
                "-cpi": subgraph.checkpoint,
                "-maxh": "0.001",
                "-noappend": None,
            },
        )

        subgraph.new = _is_new(md.output.checkpoint).output.data

        subgraph.checkpoint = md.output.checkpoint

    trajectory_continuation_loop = gmx.while_loop(
        operation=subgraph, condition=subgraph.new
    )()

    trajectory_continuation_loop.run()
    _cpt_output = trajectory_continuation_loop.output.checkpoint
    final_checkpoint = _cpt_output.result()
    loop_condition = trajectory_continuation_loop.output.new.result()
    if comm_size > 1:
        final_checkpoint = final_checkpoint[rank_number]
        loop_condition = loop_condition[rank_number]
    assert isinstance(final_checkpoint, str)
    assert isinstance(loop_condition, bool)
    assert os.path.exists(final_checkpoint)


def test_local_tools_and_assumptions():
    const = gmx.make_constant(1.0)
    assert add_float(const, const).output.data.result() == 2
    assert gmx.logical_not(less_than(const, const).output.data).result()
    # Note: It may not be safe to assume that keyword argument order (lhs, rhs) is preserved.
    assert less_than(const, add_float(const, const).output.data).output.data.result()



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
This software allows users to perform statistical test to determine if a
given molecular simulation is consistent with the thermodynamic ensemble
it is performed in.

Users should cite the JCTC paper: Shirts, M. R. "Simple Quantitative
Tests to Validate Sampling from Thermodynamic Ensembles",
J. Chem. Theory Comput., 2013, 9 (2), pp 909-926,
https://doi.org/10.1021/ct300688p
"""

import numpy as np

from .util import ensemble
from .data import SimulationData
from .util import error as pv_error


def check(data_sim_one, data_sim_two,
          total_energy=False,
          screen=False, filename=None,
          verbosity=1):
    r"""
    Check the ensemble. The correct check is inferred from the
    simulation data given.

    Parameters
    ----------
    data_sim_one : SimulationData
    data_sim_two : SimulationData
    total_energy : bool
    screen : bool
        Plot distributions on screen. Default: False.
    filename : string
        Plot distributions to `filename`.pdf. Default: None.
    verbosity : int
        Level of verbosity, from 0 (quiet) to 3 (very verbose).
        Default: 1

    Returns
    -------
    quantiles : List[float]
        The number of quantiles the computed result is off the analytical one.

    """
    if not SimulationData.compatible(data_sim_one,
                                     data_sim_two):
        raise pv_error.InputError(['data_sim_one', 'data_sim_two'],
                                  'Simulation data not compatible.')

    if data_sim_one.ensemble.ensemble != data_sim_two.ensemble.ensemble:
        raise pv_error.InputError(['data_sim_one', 'data_sim_two'],
                                  'The two simulations were sampling different ensembles. '
                                  'The simulations are expected to differ in state point '
                                  '(e.g. target temperature, target pressure), but not '
                                  'in their sampled ensemble (e.g. NVT, NPT).')

    sampled_ensemble = data_sim_one.ensemble.ensemble

    if sampled_ensemble == 'NVE' or sampled_ensemble == 'muVE':
        raise pv_error.InputError(['data_sim_one', 'data_sim_two'],
                                  'Test of ensemble ' + sampled_ensemble + ' is not implemented '
                                  '(yet).')

    if total_energy:
        eneq = 'E'
        e1 = data_sim_one.observables.total_energy
        e2 = data_sim_two.observables.total_energy
    else:
        eneq = 'U'
        e1 = data_sim_one.observables.potential_energy
        e2 = data_sim_two.observables.potential_energy

    quantiles = None

    if sampled_ensemble == 'NVT':
        quantiles = ensemble.check_1d(
            traj1=e1, traj2=e2,
            param1=data_sim_one.ensemble.temperature,
            param2=data_sim_two.ensemble.temperature,
            kb=data_sim_one.units.kb,
            quantity=eneq,
            dtemp=True, dpress=False,
            verbosity=verbosity,
            filename=filename, screen=screen
        )

    elif sampled_ensemble == 'NPT':
        temperatures = np.array([data_sim_one.ensemble.temperature,
                                 data_sim_two.ensemble.temperature])
        pressures = np.array([data_sim_one.ensemble.pressure,
                              data_sim_two.ensemble.pressure])
        equal_temps = temperatures[0] == temperatures[1]
        equal_press = pressures[0] == pressures[1]

        v1 = data_sim_one.observables.volume
        v2 = data_sim_two.observables.volume

        # Calculate conversion from p*V to energy units
        #
        # GROMACS standard units are
        #   energy: kJ/mol
        #   volume: nm^3
        #   pressure: bar
        #   => pV-term: bar * nm^3 == 1e-25 kJ == 6.022140857e-2 kJ/mol
        #   => pvconvert = 6.022140857e-2
        # UnitData stores conversion factors relative to GROMACS units
        #   energy: energy_conversion * kJ/mol
        #   volume: volume_conversion * nm^3
        #   pressure: pressure_conversion * bar
        #   => pV-term: [p]*[V] == pressure_conversion * volume_conversion bar * nm^3
        # Units were checked earlier, so we can use either simulation data structure
        pvconvert = 6.022140857e-2
        pvconvert *= (data_sim_one.units.pressure_conversion *
                      data_sim_one.units.volume_conversion)
        pvconvert /= data_sim_one.units.energy_conversion

        if equal_press and not equal_temps:
            e1 = e1 + pvconvert * pressures[0] * v1
            e2 = e2 + pvconvert * pressures[1] * v2
            if eneq == 'U':
                eneq = 'H'
            quantiles = ensemble.check_1d(
                traj1=e1, traj2=e2,
                param1=temperatures[0],
                param2=temperatures[1],
                kb=data_sim_one.units.kb,
                quantity=eneq,
                dtemp=True, dpress=False,
                verbosity=verbosity,
                filename=filename, screen=screen
            )
        elif equal_temps and not equal_press:
            quantiles = ensemble.check_1d(
                traj1=v1, traj2=v2,
                param1=pressures[0],
                param2=pressures[1],
                kb=data_sim_one.units.kb,
                quantity='V',
                dtemp=False, dpress=True,
                temp=temperatures[0],
                pvconvert=pvconvert,
                verbosity=verbosity,
                filename=filename, screen=screen
            )
        else:
            traj1 = np.array([e1, v1])
            traj2 = np.array([e2, v2])
            param1 = np.array([temperatures[0], pressures[0]])
            param2 = np.array([temperatures[1], pressures[1]])
            quantiles = ensemble.check_2d(
                traj1=traj1, traj2=traj2,
                param1=param1, param2=param2,
                kb=data_sim_one.units.kb,
                pvconvert=pvconvert,
                quantity=[eneq, 'V'],
                dtempdpress=True,
                verbosity=verbosity,
                filename=filename, screen=screen
            )

    return quantiles


def estimate_interval(data, verbosity=1, total_energy=False):
    r"""
    In order to perform an ensemble check, two simulations at distinct state
    point are needed. Choosing two state points too far apart will result
    in poor or zero overlap between the distributions, leading to very noisy
    results (due to sample errors in the tails) or a breakdown of the method,
    respectively. Choosing two state points very close to each others, on the
    other hand, makes it difficult to distinguish the slope from statistical
    error in the samples.

    This function implements a rule of thumb based on the standard deviations
    of distributions. It takes a single simulation and suggests appropriate
    intervals for a second simulation to be used for ensemble checking.

    Parameters
    ----------
    data : SimulationData
        The performed simulation.
    verbosity : int, optional
        If 0, no output is printed on screen. If 1, estimated intervals are
        printed. If larger, additional information during calculation are
        printed.
        Default: 1
    total_energy : bool, optional
        Use total energy instead of potential energy only.
        Default: False

    Returns
    -------
    intervals : Dict
        If `data` was performed under NVT conditions, `intervals` contains only
        one entry:

            * `'dT'`, containing the suggested temperature interval.

        If `data` was performed under NPT conditions, `intervals` contains three
        entries:

            * `'dT'`: Suggested temperature interval at constant pressure
            * `'dP'`: Suggested pressure interval at constant temperature
            * `'dTdP'`: Suggested combined temperature and pressure interval

    """

    if total_energy:
        ene = data.observables.total_energy
    else:
        ene = data.observables.potential_energy

    if data.ensemble.ensemble == 'NVT':
        result = ensemble.estimate_interval(
            ens_string='NVT',
            ens_temp=data.ensemble.temperature,
            energy=ene,
            kb=data.units.kb,
            verbosity=verbosity,
            tunit=data.units.temperature_str
        )
    elif data.ensemble.ensemble == 'NPT':
        pvconvert = 6.022140857e-2
        pvconvert *= (data.units.pressure_conversion *
                      data.units.volume_conversion)
        pvconvert /= data.units.energy_conversion
        result = ensemble.estimate_interval(
            ens_string='NPT',
            ens_temp=data.ensemble.temperature,
            energy=ene,
            kb=data.units.kb,
            ens_press=data.ensemble.pressure,
            volume=data.observables.volume,
            pvconvert=pvconvert,
            verbosity=verbosity,
            tunit=data.units.temperature_str,
            punit=data.units.pressure_str
        )
    else:
        raise NotImplementedError('estimate_interval() not implemented for ensemble ' +
                                  data.ensemble.ensemble)

    return result
